# Copyright 2013 GRNET S.A. All rights reserved.
#
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
#
#   1. Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and
# documentation are those of the authors and should not be
# interpreted as representing official policies, either expressed
# or implied, of GRNET S.A.

from mock import patch, call
from unittest import TestCase
from itertools import product
from json import dumps

from kamaki.clients import ClientError, cyclades

img_ref = "1m4g3-r3f3r3nc3"
vm_name = "my new VM"
fid = 42
vm_recv = dict(server=dict(
    status="BUILD",
    updated="2013-03-01T10:04:00.637152+00:00",
    hostId="",
    name=vm_name,
    imageRef=img_ref,
    created="2013-03-01T10:04:00.087324+00:00",
    flavorRef=fid,
    adminPass="n0n3sh@11p@55",
    suspended=False,
    progress=0,
    id=31173,
    metadata=dict(os="debian", users="root")))
vm_list = dict(servers=[
    dict(name='n1', id=1),
    dict(name='n2', id=2)])
net_send = dict(network=dict(dhcp=False, name='someNet'))
net_recv = dict(network=dict(
    status="PENDING",
    updated="2013-03-05T15:04:51.758780+00:00",
    name="someNet",
    created="2013-03-05T15:04:51.758728+00:00",
    cidr6=None,
    id="2130",
    gateway6=None,
    public=False,
    dhcp=False,
    cidr="192.168.1.0/24",
    type="MAC_FILTERED",
    gateway=None,
    attachments=[dict(name='att1'), dict(name='att2')]))
net_list = dict(networks=[
    dict(id=1, name='n1'),
    dict(id=2, name='n2'),
    dict(id=3, name='n3')])
firewalls = dict(attachments=[
    dict(firewallProfile='50m3_pr0f1L3', otherStuff='57uff')])


class FR(object):
    """FR stands for Fake Response"""
    json = vm_recv
    headers = {}
    content = json
    status = None
    status_code = 200

rest_pkg = 'kamaki.clients.cyclades.CycladesRestClient'
cyclades_pkg = 'kamaki.clients.cyclades.CycladesClient'


class CycladesRestClient(TestCase):

    """Set up a Cyclades thorough test"""
    def setUp(self):
        self.url = 'http://cyclades.example.com'
        self.token = 'cyc14d3s70k3n'
        self.client = cyclades.CycladesRestClient(self.url, self.token)

    def tearDown(self):
        FR.json = vm_recv

    @patch('%s.get' % rest_pkg, return_value=FR())
    def test_networks_get(self, get):
        for args in product(
                ('', 'net_id'),
                ('', 'cmd'),
                (200, 204),
                ({}, {'k': 'v'})):
            (srv_id, command, success, kwargs) = args
            self.client.networks_get(*args[:3], **kwargs)
            srv_str = '/%s' % srv_id if srv_id else ''
            cmd_str = '/%s' % command if command else ''
            self.assertEqual(get.mock_calls[-1], call(
                '/networks%s%s' % (srv_str, cmd_str),
                success=success,
                **kwargs))

    @patch('%s.delete' % rest_pkg, return_value=FR())
    def test_networks_delete(self, delete):
        for args in product(
                ('', 'net_id'),
                ('', 'cmd'),
                (202, 204),
                ({}, {'k': 'v'})):
            (srv_id, command, success, kwargs) = args
            self.client.networks_delete(*args[:3], **kwargs)
            srv_str = '/%s' % srv_id if srv_id else ''
            cmd_str = '/%s' % command if command else ''
            self.assertEqual(delete.mock_calls[-1], call(
                '/networks%s%s' % (srv_str, cmd_str),
                success=success,
                **kwargs))

    @patch('%s.set_header' % rest_pkg)
    @patch('%s.post' % rest_pkg, return_value=FR())
    def test_networks_post(self, post, SH):
        for args in product(
                ('', 'net_id'),
                ('', 'cmd'),
                (None, [dict(json="data"), dict(data="json")]),
                (202, 204),
                ({}, {'k': 'v'})):
            (srv_id, command, json_data, success, kwargs) = args
            self.client.networks_post(*args[:4], **kwargs)
            vm_str = '/%s' % srv_id if srv_id else ''
            cmd_str = '/%s' % command if command else ''
            if json_data:
                json_data = dumps(json_data)
                self.assertEqual(SH.mock_calls[-2:], [
                    call('Content-Type', 'application/json'),
                    call('Content-Length', len(json_data))])
            self.assertEqual(post.mock_calls[-1], call(
                '/networks%s%s' % (vm_str, cmd_str),
                data=json_data, success=success,
                **kwargs))

    @patch('%s.set_header' % rest_pkg)
    @patch('%s.put' % rest_pkg, return_value=FR())
    def test_networks_put(self, put, SH):
        for args in product(
                ('', 'net_id'),
                ('', 'cmd'),
                (None, [dict(json="data"), dict(data="json")]),
                (202, 204),
                ({}, {'k': 'v'})):
            (srv_id, command, json_data, success, kwargs) = args
            self.client.networks_put(*args[:4], **kwargs)
            vm_str = '/%s' % srv_id if srv_id else ''
            cmd_str = '/%s' % command if command else ''
            if json_data:
                json_data = dumps(json_data)
                self.assertEqual(SH.mock_calls[-2:], [
                    call('Content-Type', 'application/json'),
                    call('Content-Length', len(json_data))])
            self.assertEqual(put.mock_calls[-1], call(
                '/networks%s%s' % (vm_str, cmd_str),
                data=json_data, success=success,
                **kwargs))

    @patch('%s.get' % rest_pkg, return_value=FR())
    def test_floating_ip_pools_get(self, get):
        for args in product(
                (200, 204),
                ({}, {'k': 'v'})):
            success, kwargs = args
            r = self.client.floating_ip_pools_get(success, **kwargs)
            self.assertTrue(isinstance(r, FR))
            self.assertEqual(get.mock_calls[-1], call(
                '/os-floating-ip-pools', success=success, **kwargs))

    @patch('%s.get' % rest_pkg, return_value=FR())
    def test_floating_ips_get(self, get):
        for args in product(
                ('fip', ''),
                (200, 204),
                ({}, {'k': 'v'})):
            fip, success, kwargs = args
            r = self.client.floating_ips_get(fip, success, **kwargs)
            self.assertTrue(isinstance(r, FR))
            expected = '' if not fip else '/%s' % fip
            self.assertEqual(get.mock_calls[-1], call(
                '/os-floating-ips%s' % expected, success=success, **kwargs))

    @patch('%s.set_header' % rest_pkg)
    @patch('%s.post' % rest_pkg, return_value=FR())
    def test_floating_ips_post(self, post, SH):
        for args in product(
                (None, [dict(json="data"), dict(data="json")]),
                ('fip', ''),
                (202, 204),
                ({}, {'k': 'v'})):
            json_data, fip, success, kwargs = args
            self.client.floating_ips_post(*args[:3], **kwargs)
            if json_data:
                json_data = dumps(json_data)
                self.assertEqual(SH.mock_calls[-2:], [
                    call('Content-Type', 'application/json'),
                    call('Content-Length', len(json_data))])
            expected = '' if not fip else '/%s' % fip
            self.assertEqual(post.mock_calls[-1], call(
                '/os-floating-ips%s' % expected,
                data=json_data, success=success,
                **kwargs))

    @patch('%s.delete' % rest_pkg, return_value=FR())
    def test_floating_ips_delete(self, delete):
        for args in product(
                ('fip1', 'fip2'),
                (200, 204),
                ({}, {'k': 'v'})):
            fip, success, kwargs = args
            r = self.client.floating_ips_delete(fip, success, **kwargs)
            self.assertTrue(isinstance(r, FR))
            self.assertEqual(delete.mock_calls[-1], call(
                '/os-floating-ips/%s' % fip, success=success, **kwargs))


class CycladesClient(TestCase):

    def assert_dicts_are_equal(self, d1, d2):
        for k, v in d1.items():
            self.assertTrue(k in d2)
            if isinstance(v, dict):
                self.assert_dicts_are_equal(v, d2[k])
            else:
                self.assertEqual(unicode(v), unicode(d2[k]))

    """Set up a Cyclades thorough test"""
    def setUp(self):
        self.url = 'http://cyclades.example.com'
        self.token = 'cyc14d3s70k3n'
        self.client = cyclades.CycladesClient(self.url, self.token)

    def tearDown(self):
        FR.status_code = 200
        FR.json = vm_recv

    @patch('%s.servers_action_post' % cyclades_pkg, return_value=FR())
    def test_shutdown_server(self, SP):
        vm_id = vm_recv['server']['id']
        self.client.shutdown_server(vm_id)
        SP.assert_called_once_with(
            vm_id, json_data=dict(shutdown=dict()), success=202)

    @patch('%s.servers_action_post' % cyclades_pkg, return_value=FR())
    def test_start_server(self, SP):
        vm_id = vm_recv['server']['id']
        self.client.start_server(vm_id)
        SP.assert_called_once_with(
            vm_id, json_data=dict(start=dict()), success=202)

    @patch('%s.servers_action_post' % cyclades_pkg, return_value=FR())
    def test_get_server_console(self, SP):
        cnsl = dict(console=dict(info1='i1', info2='i2', info3='i3'))
        FR.json = cnsl
        vm_id = vm_recv['server']['id']
        r = self.client.get_server_console(vm_id)
        SP.assert_called_once_with(
            vm_id, json_data=dict(console=dict(type='vnc')), success=200)
        self.assert_dicts_are_equal(r, cnsl['console'])

    def test_get_firewall_profile(self):
        vm_id = vm_recv['server']['id']
        v = firewalls['attachments'][0]['firewallProfile']
        with patch.object(
                cyclades.CycladesClient, 'get_server_details',
                return_value=firewalls) as GSD:
            r = self.client.get_firewall_profile(vm_id)
            GSD.assert_called_once_with(vm_id)
            self.assertEqual(r, v)
        with patch.object(
                cyclades.CycladesClient, 'get_server_details',
                return_value=dict()):
            self.assertRaises(
                ClientError,
                self.client.get_firewall_profile,
                vm_id)

    @patch('%s.servers_action_post' % cyclades_pkg, return_value=FR())
    def test_set_firewall_profile(self, SP):
        vm_id = vm_recv['server']['id']
        v = firewalls['attachments'][0]['firewallProfile']
        self.client.set_firewall_profile(vm_id, v)
        SP.assert_called_once_with(vm_id, json_data=dict(
            firewallProfile=dict(profile=v)), success=202)

    @patch('%s.networks_post' % cyclades_pkg, return_value=FR())
    def test_create_network(self, NP):
        net_name = net_send['network']['name']
        FR.json = net_recv
        full_args = dict(
                cidr='192.168.0.0/24',
                gateway='192.168.0.1',
                type='MAC_FILTERED',
                dhcp=True)
        test_args = dict(full_args)
        test_args.update(dict(empty=None, full=None))
        net_exp = dict(dhcp=False, name=net_name, type='MAC_FILTERED')
        for arg, val in test_args.items():
            kwargs = {} if arg == 'empty' else full_args if (
                arg == 'full') else {arg: val}
            expected = dict(network=dict(net_exp))
            expected['network'].update(kwargs)
            r = self.client.create_network(net_name, **kwargs)
            self.assertEqual(
                NP.mock_calls[-1],
                call(json_data=expected, success=202))
            self.assert_dicts_are_equal(r, net_recv['network'])

    @patch('%s.networks_post' % cyclades_pkg, return_value=FR())
    def test_connect_server(self, NP):
        vm_id = vm_recv['server']['id']
        net_id = net_recv['network']['id']
        self.client.connect_server(vm_id, net_id)
        NP.assert_called_once_with(
            net_id, 'action',
            json_data=dict(add=dict(serverRef=vm_id)))

    @patch('%s.networks_post' % cyclades_pkg, return_value=FR())
    def test_disconnect_server(self, NP):
        net_id, vm_id = net_recv['network']['id'], vm_recv['server']['id']
        nic_id = 'nic-%s-%s' % (net_id, vm_id)
        vm_nics = [
            dict(id=nic_id, network_id=net_id),
            dict(id='another-nic-id', network_id='another-net-id'),
            dict(id=nic_id * 2, network_id=net_id * 2)]
        with patch.object(
                cyclades.CycladesClient,
                'list_server_nics',
                return_value=vm_nics) as LSN:
            r = self.client.disconnect_server(vm_id, nic_id)
            LSN.assert_called_once_with(vm_id)
            NP.assert_called_once_with(
                net_id, 'action',
                json_data=dict(remove=dict(attachment=nic_id)))
            self.assertEqual(r, 1)

    @patch('%s.servers_ips_get' % cyclades_pkg, return_value=FR())
    def test_list_server_nics(self, SG):
        vm_id = vm_recv['server']['id']
        nics = dict(attachments=[dict(id='nic1'), dict(id='nic2')])
        FR.json = nics
        r = self.client.list_server_nics(vm_id)
        SG.assert_called_once_with(vm_id)
        expected = nics['attachments']
        for i in range(len(r)):
            self.assert_dicts_are_equal(r[i], expected[i])
        self.assertEqual(i + 1, len(r))

    @patch('%s.networks_get' % cyclades_pkg, return_value=FR())
    def test_list_networks(self, NG):
        FR.json = net_list
        expected = net_list['networks']
        for detail in ('', 'detail'):
            r = self.client.list_networks(detail=True if detail else False)
            self.assertEqual(NG.mock_calls[-1], call(command=detail))
            for i, net in enumerate(expected):
                self.assert_dicts_are_equal(r[i], net)
            self.assertEqual(i + 1, len(r))

    @patch('%s.networks_get' % cyclades_pkg, return_value=FR())
    def test_list_network_nics(self, NG):
        net_id = net_recv['network']['id']
        FR.json = net_recv
        r = self.client.list_network_nics(net_id)
        NG.assert_called_once_with(network_id=net_id)
        expected = net_recv['network']['attachments']
        for i in range(len(r)):
            self.assert_dicts_are_equal(r[i], expected[i])

    @patch('%s.networks_post' % cyclades_pkg, return_value=FR())
    def test_disconnect_network_nics(self, NP):
        net_id = net_recv['network']['id']
        nics = ['nic1', 'nic2', 'nic3']
        with patch.object(
                cyclades.CycladesClient,
                'list_network_nics',
                return_value=nics) as LNN:
            self.client.disconnect_network_nics(net_id)
            LNN.assert_called_once_with(net_id)
            for i in range(len(nics)):
                expected = call(net_id, 'action', json_data=dict(
                    remove=dict(attachment=nics[i])))
                self.assertEqual(expected, NP.mock_calls[i])

    @patch('%s.networks_get' % cyclades_pkg, return_value=FR())
    def test_get_network_details(self, NG):
        FR.json = net_recv
        net_id = net_recv['network']['id']
        r = self.client.get_network_details(net_id)
        NG.assert_called_once_with(network_id=net_id)
        self.assert_dicts_are_equal(r, net_recv['network'])

    @patch('%s.networks_put' % cyclades_pkg, return_value=FR())
    def test_update_network_name(self, NP):
        net_id = net_recv['network']['id']
        new_name = '%s_new' % net_id
        self.client.update_network_name(net_id, new_name)
        NP.assert_called_once_with(
            network_id=net_id,
            json_data=dict(network=dict(name=new_name)))

    def test_delete_network(self):
        net_id = net_recv['network']['id']
        with patch.object(
                cyclades.CycladesClient, 'networks_delete',
                return_value=FR()) as ND:
            self.client.delete_network(net_id)
            ND.assert_called_once_with(net_id)
        with patch.object(
                cyclades.CycladesClient, 'networks_delete',
                side_effect=ClientError('A 421 Error', 421)):
            try:
                self.client.delete_network(421)
            except ClientError as err:
                self.assertEqual(err.status, 421)
                self.assertEqual(err.details, [
                    'Network may be still connected to at least one server'])

    @patch('%s.floating_ip_pools_get' % cyclades_pkg, return_value=FR())
    def test_get_floating_ip_pools(self, get):
        r = self.client.get_floating_ip_pools()
        self.assert_dicts_are_equal(r, FR.json)
        self.assertEqual(get.mock_calls[-1], call())

    @patch('%s.floating_ips_get' % cyclades_pkg, return_value=FR())
    def test_get_floating_ips(self, get):
        r = self.client.get_floating_ips()
        self.assert_dicts_are_equal(r, FR.json)
        self.assertEqual(get.mock_calls[-1], call())

    @patch('%s.floating_ips_post' % cyclades_pkg, return_value=FR())
    def test_alloc_floating_ip(self, post):
        FR.json = dict(floating_ip=dict(
            fixed_ip='fip',
            id=1,
            instance_id='lala',
            ip='102.0.0.1',
            pool='pisine'))
        for args in product(
                (None, 'pisine'),
                (None, 'Iwannanip')):
            r = self.client.alloc_floating_ip(*args)
            pool, address = args
            self.assert_dicts_are_equal(r, FR.json['floating_ip'])
            json_data = dict()
            if pool:
                json_data['pool'] = pool
            if address:
                json_data['address'] = address
            self.assertEqual(post.mock_calls[-1], call(json_data))

    @patch('%s.floating_ips_get' % cyclades_pkg, return_value=FR())
    def test_get_floating_ip(self, get):
        FR.json = dict(floating_ip=dict(
            fixed_ip='fip',
            id=1,
            instance_id='lala',
            ip='102.0.0.1',
            pool='pisine'))
        self.assertRaises(AssertionError, self.client.get_floating_ip, None)
        fip = 'fip'
        r = self.client.get_floating_ip(fip)
        self.assert_dicts_are_equal(r, FR.json['floating_ip'])
        self.assertEqual(get.mock_calls[-1], call(fip))

    @patch('%s.floating_ips_delete' % cyclades_pkg, return_value=FR())
    def test_delete_floating_ip(self, delete):
        self.assertRaises(AssertionError, self.client.delete_floating_ip, None)
        fip = 'fip'
        r = self.client.delete_floating_ip(fip)
        self.assert_dicts_are_equal(r, FR.headers)
        self.assertEqual(delete.mock_calls[-1], call(fip))

    @patch('%s.servers_action_post' % cyclades_pkg, return_value=FR())
    def test_attach_floating_ip(self, spost):
        vmid, addr = 42, 'anIpAddress'
        for err, args in {
                ValueError: ['not a server id', addr],
                TypeError: [None, addr],
                AssertionError: [vmid, None],
                AssertionError: [vmid, '']}.items():
            self.assertRaises(
                err, self.client.attach_floating_ip, *args)
        r = self.client.attach_floating_ip(vmid, addr)
        self.assert_dicts_are_equal(r, FR.headers)
        expected = dict(addFloatingIp=dict(address=addr))
        self.assertEqual(spost.mock_calls[-1], call(vmid, json_data=expected))

    @patch('%s.servers_action_post' % cyclades_pkg, return_value=FR())
    def test_detach_floating_ip(self, spost):
        vmid, addr = 42, 'anIpAddress'
        for err, args in {
                ValueError: ['not a server id', addr],
                TypeError: [None, addr],
                AssertionError: [vmid, None],
                AssertionError: [vmid, '']}.items():
            self.assertRaises(
                err, self.client.detach_floating_ip, *args)
        r = self.client.detach_floating_ip(vmid, addr)
        self.assert_dicts_are_equal(r, FR.headers)
        expected = dict(removeFloatingIp=dict(address=addr))
        self.assertEqual(spost.mock_calls[-1], call(vmid, json_data=expected))


if __name__ == '__main__':
    from sys import argv
    from kamaki.clients.test import runTestCase
    not_found = True
    if not argv[1:] or argv[1] == 'CycladesClient':
        not_found = False
        runTestCase(CycladesClient, 'Cyclades Client', argv[2:])
    if not argv[1:] or argv[1] == 'CycladesRestClient':
        not_found = False
        runTestCase(CycladesRestClient, 'CycladesRest Client', argv[2:])
    if not_found:
        print('TestCase %s not found' % argv[1])



# Copyright 2013 GRNET S.A. All rights reserved.
#
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
#
#   1. Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and
# documentation are those of the authors and should not be
# interpreted as representing official policies, either expressed
# or implied, of GRNET S.A.command

from json import loads, load
from sys import stdin
from os.path import abspath

from astakosclient import AstakosClient, AstakosClientException

from kamaki.cli import command
from kamaki.cli.errors import CLIBaseUrlError
from kamaki.cli.commands import (
    _command_init, errors, _optional_json, addLogSettings)
from kamaki.cli.command_tree import CommandTree
from kamaki.cli.utils import format_size
from kamaki.cli.argument import FlagArgument, ValueArgument, IntArgument
from kamaki.cli.argument import CommaSeparatedListArgument
from kamaki.cli.logger import get_logger

snfastakos_cmds = CommandTree('astakos', 'astakosclient CLI')
snfproject_cmds = CommandTree('project', 'Synnefo project management CLI')
_commands = [snfastakos_cmds, snfproject_cmds]


def astakoserror(foo):
    def _raise(self, *args, **kwargs):
        try:
            return foo(self, *args, **kwargs)
        except AstakosClientException as ace:
            try:
                ace.details = ['%s' % ace.details]
            except Exception:
                pass
            finally:
                raise ace
    return _raise


class _astakos_init(_command_init):

    def __init__(self, arguments=dict(), auth_base=None, cloud=None):
        super(_astakos_init, self).__init__(arguments, auth_base, cloud)
        self['token'] = ValueArgument('Custom token', '--token')

    @errors.generic.all
    @astakoserror
    @addLogSettings
    def _run(self):
        self.cloud = self.cloud if self.cloud else 'default'
        self.token = self['token'] or self._custom_token('astakos')\
            or self.config.get_cloud(self.cloud, 'token')
        if getattr(self, 'auth_base', False):
            astakos_endpoints = self.auth_base.get_service_endpoints(
                self._custom_type('astakos') or 'identity',
                self._custom_version('astakos') or '')
            base_url = astakos_endpoints['publicURL']
            base_url, sep, suffix = base_url.rpartition('identity')
        else:
            base_url = self._custom_url('astakos')
        if not base_url:
            raise CLIBaseUrlError(service='astakos')
        self.client = AstakosClient(
            base_url, logger=get_logger('kamaki.clients'))

    def main(self):
        self._run()


@command(snfastakos_cmds)
class astakos_user_info(_astakos_init, _optional_json):
    """Authenticate a user
    Get user information (e.g., unique account name) from token
    Token should be set in settings:
    *  check if a token is set    /config get cloud.default.token
    *  permanently set a token    /config set cloud.default.token <token>
    Token can also be provided as a parameter
    (To use a named cloud, use its name instead of "default")
    """

    arguments = dict(
        usage=FlagArgument('also return usage information', ('--with-usage'))
    )

    @errors.generic.all
    @astakoserror
    def _run(self):
        self._print(
            self.client.get_user_info(self.token, self['usage']),
            self.print_dict)

    def main(self):
        super(self.__class__, self)._run()
        self._run()


@command(snfastakos_cmds)
class astakos_user_name(_astakos_init, _optional_json):
    """Get username(s) from uuid(s)"""

    arguments = dict(
        service_token=ValueArgument(
            'Use service token instead', '--service-token')
    )

    @errors.generic.all
    @astakoserror
    def _run(self, uuids):
        assert uuids and isinstance(uuids, list), 'No valid uuids'
        if 1 == len(uuids):
            self._print(self.client.get_username(self.token, uuids[0]))
        else:
            self._print(
                self.client.get_usernames(self.token, uuids), self.print_dict)

    def main(self, uuid, *more_uuids):
        super(self.__class__, self)._run()
        self._run([uuid] + list(more_uuids))


@command(snfastakos_cmds)
class astakos_user_uuid(_astakos_init, _optional_json):
    """Get uuid(s) from username(s)"""

    @errors.generic.all
    @astakoserror
    def _run(self, usernames):
        assert usernames and isinstance(usernames, list), 'No valid usernames'
        if 1 == len(usernames):
            self._print(self.client.get_uuid(self.token, usernames[0]))
        else:
            self._print(
                self.client.get_uuids(self.token, usernames), self.print_dict)

    def main(self, usernames, *more_usernames):
        super(self.__class__, self)._run()
        self._run([usernames] + list(more_usernames))


@command(snfastakos_cmds)
class astakos_quotas(_astakos_init, _optional_json):
    """Get user (or service) quotas"""

    def _print_with_format(self, d, out):
        """ Print d with size formating when needed
        :param d: (dict) {system: {<service>: {usage: ..., limit: ..., }, ...}}
        """
        newd = dict()
        for k, service in d['system'].items():
            newd[k] = dict(service)
            for term in ('usage', 'limit'):
                if term in service:
                    newd[k][term] = format_size(service[term])
        self.print_dict(newd, out=out)

    @errors.generic.all
    @astakoserror
    def _run(self):
            self._print(
                self.client.get_quotas(self.token), self._print_with_format)

    def main(self):
        super(self.__class__, self)._run()
        self._run()


@command(snfastakos_cmds)
class astakos_services(_astakos_init):
    """Astakos operations filtered by services"""


@command(snfastakos_cmds)
class astakos_services_list(_astakos_init, _optional_json):
    """List available services"""

    @errors.generic.all
    @astakoserror
    def _run(self):
        self._print(self.client.get_services())

    def main(self):
        super(self.__class__, self)._run()
        self._run()


@command(snfastakos_cmds)
class astakos_services_username(_astakos_init, _optional_json):
    """Get service username(s) from uuid(s)"""

    @errors.generic.all
    @astakoserror
    def _run(self, stoken, uuids):
        assert uuids and isinstance(uuids, list), 'No valid uuids'
        if 1 == len(uuids):
            self._print(self.client.service_get_username(stoken, uuids[0]))
        else:
            self._print(
                self.client.service_get_usernames(stoken, uuids),
                self.print_dict)

    def main(self, service_token, uuid, *more_uuids):
        super(self.__class__, self)._run()
        self._run(service_token, [uuid] + list(more_uuids))


@command(snfastakos_cmds)
class astakos_services_uuid(_astakos_init, _optional_json):
    """Get service uuid(s) from username(s)"""

    @errors.generic.all
    @astakoserror
    def _run(self, stoken, usernames):
        assert usernames and isinstance(usernames, list), 'No valid usernames'
        if 1 == len(usernames):
            self._print(self.client.service_get_uuid(self.token, usernames[0]))
        else:
            self._print(
                self.client.service_get_uuids(self.token, usernames),
                self.print_dict)

    def main(self, service_token, usernames, *more_usernames):
        super(self.__class__, self)._run()
        self._run(service_token, [usernames] + list(more_usernames))


@command(snfastakos_cmds)
class astakos_services_quotas(_astakos_init, _optional_json):
    """Get user (or service) quotas"""

    arguments = dict(
        uuid=ValueArgument('A user unique id to get quotas for', '--uuid')
    )

    @errors.generic.all
    @astakoserror
    def _run(self, stoken):
        self._print(self.client.service_get_quotas(stoken, self['uuid']))

    def main(self, service_token):
        super(self.__class__, self)._run()
        self._run(service_token)


@command(snfastakos_cmds)
class astakos_resources(_astakos_init, _optional_json):
    """List user resources"""

    @errors.generic.all
    @astakoserror
    def _run(self):
        self._print(self.client.get_resources(), self.print_dict)

    def main(self):
        super(self.__class__, self)._run()
        self._run()


@command(snfastakos_cmds)
class astakos_feedback(_astakos_init):
    """Send feedback to astakos server"""

    @errors.generic.all
    @astakoserror
    def _run(self, msg, more_info=None):
        self.client.send_feedback(self.token, msg, more_info or '')

    def main(self, message, more_info=None):
        super(self.__class__, self)._run()
        self._run(message, more_info)


@command(snfastakos_cmds)
class astakos_endpoints(_astakos_init, _optional_json):
    """Get endpoints service endpoints"""

    arguments = dict(uuid=ValueArgument('User uuid', '--uuid'))

    @errors.generic.all
    @astakoserror
    def _run(self):
        self._print(
            self.client.get_endpoints(self.token, self['uuid']),
            self.print_dict)

    def main(self):
        super(self.__class__, self)._run()
        self._run()


@command(snfastakos_cmds)
class astakos_commission(_astakos_init):
    """Manage commissions (special privileges required)"""


@command(snfastakos_cmds)
class astakos_commission_pending(_astakos_init, _optional_json):
    """List pending commissions (special privileges required)"""

    @errors.generic.all
    @astakoserror
    def _run(self):
        self._print(self.client.get_pending_commissions(self.token))

    def main(self):
        super(self.__class__, self)._run()
        self._run()


@command(snfastakos_cmds)
class astakos_commission_info(_astakos_init, _optional_json):
    """Get commission info (special privileges required)"""

    @errors.generic.all
    @astakoserror
    def _run(self, commission_id):
        commission_id = int(commission_id)
        self._print(
            self.client.get_commission_info(self.token, commission_id),
            self.print_dict)

    def main(self, commission_id):
        super(self.__class__, self)._run()
        self._run(commission_id)


@command(snfastakos_cmds)
class astakos_commission_action(_astakos_init, _optional_json):
    """Invoke an action in a commission (special privileges required)
    Actions can be accept or reject
    """

    actions = ('accept', 'reject')

    @errors.generic.all
    @astakoserror
    def _run(self, commission_id, action):
        commission_id = int(commission_id)
        action = action.lower()
        assert action in self.actions, 'Actions can be %s' % (
            ' or '.join(self.actions))
        self._print(
            self.client.commission_acction(self.token, commission_id, action),
            self.print_dict)

    def main(self, commission_id, action):
        super(self.__class__, self)._run()
        self._run(commission_id, action)


@command(snfastakos_cmds)
class astakos_commission_accept(_astakos_init):
    """Accept a pending commission  (special privileges required)"""

    @errors.generic.all
    @astakoserror
    def _run(self, commission_id):
        commission_id = int(commission_id)
        self.client.accept_commission(self.token, commission_id)

    def main(self, commission_id):
        super(self.__class__, self)._run()
        self._run(commission_id)


@command(snfastakos_cmds)
class astakos_commission_reject(_astakos_init):
    """Reject a pending commission  (special privileges required)"""

    @errors.generic.all
    @astakoserror
    def _run(self, commission_id):
        commission_id = int(commission_id)
        self.client.reject_commission(self.token, commission_id)

    def main(self, commission_id):
        super(self.__class__, self)._run()
        self._run(commission_id)


@command(snfastakos_cmds)
class astakos_commission_resolve(_astakos_init, _optional_json):
    """Resolve multiple commissions  (special privileges required)"""

    arguments = dict(
        accept=CommaSeparatedListArgument(
            'commission ids to accept (e.g., --accept=11,12,13,...',
            '--accept'),
        reject=CommaSeparatedListArgument(
            'commission ids to reject (e.g., --reject=11,12,13,...',
            '--reject'),
    )

    @errors.generic.all
    @astakoserror
    def _run(self):
        self.writeln('accepted ', self['accept'])
        self.writeln('rejected ', self['reject'])
        self._print(
            self.client.resolve_commissions(
                self.token, self['accept'], self['reject']),
            self.print_dict)

    def main(self):
        super(self.__class__, self)._run()
        self._run()


@command(snfastakos_cmds)
class astakos_commission_issue(_astakos_init, _optional_json):
    """Issue commissions as a json string (special privileges required)
    Parameters:
    holder      -- user's id (string)
    source      -- commission's source (ex system) (string)
    provisions  -- resources with their quantity (json-dict from string to int)
    name        -- description of the commission (string)
    """

    arguments = dict(
        force=FlagArgument('Force commission', '--force'),
        accept=FlagArgument('Do not wait for verification', '--accept')
    )

    @errors.generic.all
    @astakoserror
    def _run(
            self, holder, source, provisions, name=''):
        provisions = loads(provisions)
        self._print(self.client.issue_one_commission(
            self.token, holder, source, provisions, name,
            self['force'], self['accept']))

    def main(self, holder, source, provisions, name=''):
        super(self.__class__, self)._run()
        self._run(holder, source, provisions, name)


@command(snfastakos_cmds)
class astakos_commission_issuejson(_astakos_init, _optional_json):
    """Issue commissions as a json string (special privileges required)"""

    @errors.generic.all
    @astakoserror
    def _run(self, info_json):
        infodict = loads(info_json)
        self._print(self.client.issue_commission(self.token, infodict))

    def main(self, info_json):
        super(self.__class__, self)._run()
        self._run(info_json)

# XXX issue_commission, issue_one_commission


# Project commands


_project_specs = """
    {
        "name": name,
        "owner": uuid,
        "homepage": homepage,         # optional
        "description": description,   # optional
        "comments": comments,         # optional
        "start_date": date,           # optional
        "end_date": date,
        "join_policy": "auto" | "moderated" | "closed",  # default: "moderated"
        "leave_policy": "auto" | "moderated" | "closed", # default: "auto"
        "resources": {
            "cyclades.vm": {
                "project_capacity": int or null,
                 "member_capacity": int
            }
        }
  }
  """


def apply_notification(foo):
    def wrap(self, *args, **kwargs):
        r = foo(self, *args, **kwargs)
        self.writeln('Application is submitted successfully')
        return r
    return wrap


@command(snfproject_cmds)
class project_list(_astakos_init, _optional_json):
    """List all projects"""

    arguments = dict(
        name=ValueArgument('Filter by name', ('--with-name', )),
        state=ValueArgument('Filter by state', ('--with-state', )),
        owner=ValueArgument('Filter by owner', ('--with-owner', ))
    )

    @errors.generic.all
    @astakoserror
    def _run(self):
        self._print(self.client.get_projects(
            self.token, self['name'], self['state'], self['owner']))

    def main(self):
        super(self.__class__, self)._run()
        self._run()


@command(snfproject_cmds)
class project_info(_astakos_init, _optional_json):
    """Get details for a project"""

    @errors.generic.all
    @astakoserror
    def _run(self, project_id):
        self._print(
            self.client.get_project(self.token, project_id), self.print_dict)

    def main(self, project_id):
        super(self.__class__, self)._run()
        self._run(project_id)


@command(snfproject_cmds)
class project_create(_astakos_init, _optional_json):
    """Apply for a new project (input a json-dict)
    Project details must be provided as a json-formated dict from the standard
    input, or through a file
    """

    __doc__ += _project_specs

    arguments = dict(
        specs_path=ValueArgument(
            'Specification file path (content must be in json)', '--spec-file')
    )

    @errors.generic.all
    @astakoserror
    @apply_notification
    def _run(self):
        input_stream = open(abspath(self['specs_path'])) if (
            self['specs_path']) else stdin
        specs = load(input_stream)
        self._print(
            self.client.create_project(self.token, specs), self.print_dict)

    def main(self):
        super(self.__class__, self)._run()
        self._run()


@command(snfproject_cmds)
class project_modify(_astakos_init, _optional_json):
    """Modify a project (input a json-dict)
    Project details must be provided as a json-formated dict from the standard
    input, or through a file
    """

    __doc__ += _project_specs

    arguments = dict(
        specs_path=ValueArgument(
            'Specification file path (content must be in json)', '--spec-file')
    )

    @errors.generic.all
    @astakoserror
    @apply_notification
    def _run(self, project_id):
        input_stream = open(abspath(self['specs_path'])) if (
            self['specs_path']) else stdin
        specs = load(input_stream)
        self._print(
            self.client.modify_project(self.token, project_id, specs),
            self.print_dict)

    def main(self, project_id):
        super(self.__class__, self)._run()
        self._run(project_id)


class _project_action(_astakos_init):

    action = ''

    @errors.generic.all
    @astakoserror
    def _run(self, project_id, quote_a_reason):
        self.client.project_action(
            self.token, project_id, self.action, quote_a_reason)

    def main(self, project_id, quote_a_reason=''):
        super(_project_action, self)._run()
        self._run(project_id, quote_a_reason)


@command(snfproject_cmds)
class project_suspend(_project_action):
    """Suspend a project (special privileges needed)"""
    action = 'suspend'


@command(snfproject_cmds)
class project_unsuspend(_project_action):
    """Resume a suspended project (special privileges needed)"""
    action = 'unsuspend'


@command(snfproject_cmds)
class project_terminate(_project_action):
    """Terminate a project (special privileges needed)"""
    action = 'terminate'


@command(snfproject_cmds)
class project_reinstate(_project_action):
    """Reinstate a terminated project (special privileges needed)"""
    action = 'reinstate'


@command(snfproject_cmds)
class project_application(_astakos_init):
    """Application management commands"""


@command(snfproject_cmds)
class project_application_list(_astakos_init, _optional_json):
    """List all applications (old and new)"""

    arguments = dict(
        project=IntArgument('Filter by project id', '--with-project-id')
    )

    @errors.generic.all
    @astakoserror
    def _run(self):
        self._print(self.client.get_applications(self.token, self['project']))

    def main(self):
        super(self.__class__, self)._run()
        self._run()


@command(snfproject_cmds)
class project_application_info(_astakos_init, _optional_json):
    """Get details on an application"""

    @errors.generic.all
    @astakoserror
    def _run(self, app_id):
        self._print(
            self.client.get_application(self.token, app_id), self.print_dict)

    def main(self, application_id):
        super(self.__class__, self)._run()
        self._run(application_id)


class _application_action(_astakos_init):

    action = ''

    @errors.generic.all
    @astakoserror
    def _run(self, app_id, quote_a_reason):
        self.client.application_action(
            self.token, app_id, self.action, quote_a_reason)

    def main(self, application_id, quote_a_reason=''):
        super(_application_action, self)._run()
        self._run(application_id, quote_a_reason)


@command(snfproject_cmds)
class project_application_approve(_application_action):
    """Approve an application (special privileges needed)"""
    action = 'approve'


@command(snfproject_cmds)
class project_application_deny(_application_action):
    """Deny an application (special privileges needed)"""
    action = 'deny'


@command(snfproject_cmds)
class project_application_dismiss(_application_action):
    """Dismiss your denied application"""
    action = 'dismiss'


@command(snfproject_cmds)
class project_application_cancel(_application_action):
    """Cancel your application"""
    action = 'cancel'


@command(snfproject_cmds)
class project_membership(_astakos_init):
    """Project membership management commands"""


@command(snfproject_cmds)
class project_membership_list(_astakos_init, _optional_json):
    """List all memberships"""

    arguments = dict(
        project=IntArgument('Filter by project id', '--with-project-id')
    )

    @errors.generic.all
    @astakoserror
    def _run(self):
        self._print(self.client.get_memberships(self.token, self['project']))

    def main(self):
        super(self.__class__, self)._run()
        self._run()


@command(snfproject_cmds)
class project_membership_info(_astakos_init, _optional_json):
    """Details on a membership"""

    @errors.generic.all
    @astakoserror
    def _run(self, memb_id):
        self._print(
            self.client.get_membership(self.token, memb_id), self.print_dict)

    def main(self, membership_id):
        super(self.__class__, self)._run()
        self._run(membership_id)


class _membership_action(_astakos_init, _optional_json):

    action = ''

    @errors.generic.all
    @astakoserror
    def _run(self, memb_id, quote_a_reason):
        self._print(self.client.membership_action(
            self.token, memb_id, self.action, quote_a_reason))

    def main(self, membership_id, quote_a_reason=''):
        super(_membership_action, self)._run()
        self._run(membership_id, quote_a_reason)


@command(snfproject_cmds)
class project_membership_leave(_membership_action):
    """Leave a project you have membership to"""
    action = 'leave'


@command(snfproject_cmds)
class project_membership_cancel(_membership_action):
    """Cancel your (probably pending) membership to a project"""
    action = 'cancel'


@command(snfproject_cmds)
class project_membership_accept(_membership_action):
    """Accept a membership for a project you manage"""
    action = 'accept'


@command(snfproject_cmds)
class project_membership_reject(_membership_action):
    """Reject a membership for a project you manage"""
    action = 'reject'


@command(snfproject_cmds)
class project_membership_remove(_membership_action):
    """Remove a membership for a project you manage"""
    action = 'remove'


@command(snfproject_cmds)
class project_membership_join(_astakos_init):
    """Join a project"""

    @errors.generic.all
    @astakoserror
    def _run(self, project_id):
        self.writeln(self.client.join_project(self.token, project_id))

    def main(self, project_id):
        super(project_membership_join, self)._run()
        self._run(project_id)


@command(snfproject_cmds)
class project_membership_enroll(_astakos_init):
    """Enroll somebody to a project you manage"""

    @errors.generic.all
    @astakoserror
    def _run(self, project_id, email):
        self.writeln(self.client.enroll_member(self.token, project_id, email))

    def main(self, project_id, email):
        super(project_membership_join, self)._run()
        self._run(project_id, email)



# Copyright 2012-2013 GRNET S.A. All rights reserved.
#
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
#
#   1. Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and
# documentation are those of the authors and should not be
# interpreted as representing official policies, either expressed
# or implied, of GRNET S.A.

import time
import datetime
from os import urandom
from tempfile import NamedTemporaryFile
from string import ascii_letters

from kamaki.clients import livetest, ClientError
from kamaki.clients.pithos import PithosClient
from kamaki.clients.astakos import AstakosClient


def chargen():
    """10 + 2 * 26 + 26 = 88"""
    while True:
        for CH in xrange(10):
            yield '%s' % CH
        for CH in ascii_letters:
            yield CH
        for CH in '~!@#$%^&*()_+`-=:";|<>?,./':
            yield CH


def sample_block(f, block):
    block_size = 4 * 1024 * 1024
    f.seek(block * block_size)
    ch = [f.read(1)]
    f.seek(block_size / 2, 1)
    ch.append(f.read(1))
    f.seek((block + 1) * block_size - 1)
    ch.append(f.read(1))
    return ch


class Pithos(livetest.Generic):

    files = []

    def setUp(self):
        self.cloud = 'cloud.%s' % self['testcloud']
        aurl, self.token = self[self.cloud, 'url'], self[self.cloud, 'token']
        self.auth_base = AstakosClient(aurl, self.token)
        purl = self.auth_base.get_service_endpoints(
            'object-store')['publicURL']
        self.uuid = self.auth_base.user_term('id')
        self.client = PithosClient(purl, self.token, self.uuid)

        self.now = time.mktime(time.gmtime())
        self.now_unformated = datetime.datetime.utcnow()
        self._init_data()

        """Prepare an object to be shared - also its container"""
        self.client.container = self.c1
        self.client.object_post(
            'test',
            update=True,
            permissions={'read': [self.client.account]})

        self.create_remote_object(self.c1, 'another.test')

    def _init_data(self):
        self.c1 = 'c1_' + unicode(self.now)
        self.c2 = 'c2_' + unicode(self.now)
        self.c3 = 'c3_' + unicode(self.now)
        try:
            self.client.create_container(self.c2)
        except ClientError:
            pass
        try:
            self.client.create_container(self.c1)
        except ClientError:
            pass
        try:
            self.client.create_container(self.c3)
        except ClientError:
            pass

        self.create_remote_object(self.c1, 'test')
        self.create_remote_object(self.c2, 'test')
        self.create_remote_object(self.c1, 'test1')
        self.create_remote_object(self.c2, 'test1')

    def create_remote_object(self, container, obj):
        self.client.container = container
        self.client.object_put(
            obj,
            content_type='application/octet-stream',
            data='file %s that lives in %s' % (obj, container),
            metadata={'incontainer': container})

    def forceDeleteContainer(self, container):
        self.client.container = container
        try:
            r = self.client.list_objects()
        except ClientError:
            return
        for obj in r:
            name = obj['name']
            self.client.del_object(name)
        r = self.client.container_delete()
        self.container = ''

    def tearDown(self):
        """Destroy test cases"""
        for f in self.files:
            f.close()
        self.forceDeleteContainer(self.c1)
        self.forceDeleteContainer(self.c2)
        try:
            self.forceDeleteContainer(self.c3)
        except ClientError:
            pass
        self.client.container = ''

    def test_000(self):
        """Prepare a full Pithos+ test"""
        print('')
        super(self.__class__, self).test_000()

    def test_account_head(self):
        """Test account_HEAD"""
        self._test_0010_account_head()

    def _test_0010_account_head(self):
        r = self.client.account_head()
        self.assertEqual(r.status_code, 204)

        r = self.client.account_head(until='1000000000')
        self.assertEqual(r.status_code, 204)

        r = self.client.get_account_info(until='1000000000')
        datestring = unicode(r['x-account-until-timestamp'])
        self.assertEqual(u'Sun, 09 Sep 2001 01:46:40 GMT', datestring)

        r = self.client.get_account_quota()
        self.assertTrue('x-account-policy-quota' in r)

        #r = self.client.get_account_versioning()
        #self.assertTrue('x-account-policy-versioning' in r)

        """Check if(un)modified_since"""
        for format in self.client.DATE_FORMATS:
            now_formated = self.now_unformated.strftime(format)
            r1 = self.client.account_head(
                if_modified_since=now_formated, success=(204, 304, 412))
            sc1 = r1.status_code
            r2 = self.client.account_head(
                if_unmodified_since=now_formated, success=(204, 304, 412))
            sc2 = r2.status_code
            self.assertNotEqual(sc1, sc2)

    def test_account_get(self):
        """Test account_GET"""
        self._test_0020_account_get()

    def _test_0020_account_get(self):
        #r = self.client.account_get()
        #self.assertEqual(r.status_code, 200)
        r = self.client.list_containers()
        fullLen = len(r)
        self.assertTrue(fullLen > 2)

        r = self.client.account_get(limit=1)
        self.assertEqual(len(r.json), 1)

        r = self.client.account_get(marker='c2_')
        temp_c0 = r.json[0]['name']
        temp_c2 = r.json[2]['name']

        r = self.client.account_get(limit=2, marker='c2_')
        conames = [container['name'] for container in r.json if (
            container['name'].lower().startswith('c2_'))]
        self.assertTrue(temp_c0 in conames)
        self.assertFalse(temp_c2 in conames)

        r = self.client.account_get(show_only_shared=True)
        self.assertTrue(self.c1 in [c['name'] for c in r.json])

        r = self.client.account_get(until=1342609206.0)
        self.assertTrue(len(r.json) <= fullLen)

        """Check if(un)modified_since"""
        for format in self.client.DATE_FORMATS:
            now_formated = self.now_unformated.strftime(format)
            r1 = self.client.account_get(
                if_modified_since=now_formated, success=(200, 304, 412))
            sc1 = r1.status_code
            r2 = self.client.account_get(
                if_unmodified_since=now_formated, success=(200, 304, 412))
            sc2 = r2.status_code
            self.assertNotEqual(sc1, sc2)

        """Check sharing_accounts"""
        r = self.client.get_sharing_accounts()
        try:
            self.assertTrue(len(r) > 0)
        except AssertionError as e:
            print '\n\tWARNING: Are there any sharers to your account?'
            self.assertEqual(len(r), 0)
            print '\tIf there are, this (%s) is an error, else it is OK' % e

    def test_account_post(self):
        """Test account_POST"""
        self._test_0030_account_post()

    def _test_0030_account_post(self):
        r = self.client.account_post()
        self.assertEqual(r.status_code, 202)
        grpName = 'grp' + unicode(self.now)

        """Method set/del_account_meta and set_account_groupcall use
            account_post internally
        """
        u1 = self.client.account
        #  Invalid display name
        u2 = '1nc0r3c7-d15p14y-n4m3'
        self.assertRaises(
            ClientError,
            self.client.set_account_group,
            grpName, [u1, u2])
        self.client.set_account_group(grpName, [u1])
        r = self.client.get_account_group()
        self.assertEqual(r['x-account-group-' + grpName], '%s' % u1)
        self.client.del_account_group(grpName)
        r = self.client.get_account_group()
        self.assertTrue('x-account-group-' + grpName not in r)

        mprefix = 'meta' + unicode(self.now)
        self.client.set_account_meta({
            mprefix + '1': 'v1', mprefix + '2': 'v2'})
        r = self.client.get_account_meta()
        self.assertEqual(r['x-account-meta-' + mprefix + '1'], 'v1')
        self.assertEqual(r['x-account-meta-' + mprefix + '2'], 'v2')

        self.client.del_account_meta(mprefix + '1')
        r = self.client.get_account_meta()
        self.assertTrue('x-account-meta-' + mprefix + '1' not in r)

        self.client.del_account_meta(mprefix + '2')
        r = self.client.get_account_meta()
        self.assertTrue('x-account-meta-' + mprefix + '2' not in r)

        """Missing testing for quota, versioning, because normally
        you don't have permissions to modify those at account level
        """

        #newquota = 1000000
        #self.client.set_account_quota(newquota)
        #r = self.client.get_account_info()
        #print(unicode(r))
        #r = self.client.get_account_quota()
        #self.assertEqual(r['x-account-policy-quota'], newquota)
        #self.client.set_account_versioning('auto')

    def test_container_head(self):
        """Test container_HEAD"""
        self._test_0040_container_head()

    def _test_0040_container_head(self):
        self.client.container = self.c1

        r = self.client.container_head()
        self.assertEqual(r.status_code, 204)

        """Check until"""
        r = self.client.container_head(until=1000000, success=(204, 404))
        self.assertEqual(r.status_code, 404)

        """Check and if(un)modified_since"""
        for format in self.client.DATE_FORMATS:
            now_formated = self.now_unformated.strftime(format)
            r1 = self.client.container_head(
                if_modified_since=now_formated, success=(204, 304, 412))
            sc1 = r1.status_code
            r2 = self.client.container_head(
                if_unmodified_since=now_formated, success=(204, 304, 412))
            sc2 = r2.status_code
            self.assertNotEqual(sc1, sc2)

        """Check container object meta"""
        r = self.client.get_container_object_meta()
        self.assertEqual(r['x-container-object-meta'], 'Incontainer')

    def test_container_get(self):
        """Test container_GET"""
        self._test_0050_container_get()

    def _test_0050_container_get(self):
        self.client.container = self.c1

        r = self.client.container_get()
        self.assertEqual(r.status_code, 200)
        fullLen = len(r.json)

        r = self.client.container_get(prefix='test')
        lalobjects = [obj for obj in r.json if obj['name'].startswith('test')]
        self.assertTrue(len(r.json) > 1)
        self.assertEqual(len(r.json), len(lalobjects))

        r = self.client.container_get(limit=1)
        self.assertEqual(len(r.json), 1)

        r = self.client.container_get(marker='another')
        self.assertTrue(len(r.json) > 1)
        neobjects = [obj for obj in r.json if obj['name'] > 'another']
        self.assertEqual(len(r.json), len(neobjects))

        r = self.client.container_get(prefix='another.test', delimiter='.')
        self.assertTrue(fullLen > len(r.json))

        r = self.client.container_get(path='/')
        self.assertEqual(fullLen, len(r.json))

        r = self.client.container_get(format='xml')
        self.assertEqual(r.text.split()[4], 'name="' + self.c1 + '">')

        r = self.client.container_get(meta=['incontainer'])
        self.assertTrue(len(r.json) > 0)

        r = self.client.container_get(show_only_shared=True)
        self.assertTrue(len(r.json) < fullLen)

        try:
            r = self.client.container_get(until=1000000000)
            datestring = unicode(r.headers['x-account-until-timestamp'])
            self.assertEqual(u'Sun, 09 Sep 2001 01:46:40 GMT', datestring)

        except ClientError:

            pass

        """Check and if un/modified_since"""
        for format in self.client.DATE_FORMATS:
            now_formated = self.now_unformated.strftime(format)
            r1 = self.client.container_get(
                if_modified_since=now_formated, success=(200, 304, 412))
            sc1 = r1.status_code
            r2 = self.client.container_get(
                if_unmodified_since=now_formated, success=(200, 304, 412))
            sc2 = r2.status_code
            self.assertNotEqual(sc1, sc2)

    def test_container_put(self):
        """Test container_PUT"""
        self._test_0050_container_put()

    def _test_0050_container_put(self):
        self.client.container = self.c2

        r = self.client.create_container()
        self.assertTrue(isinstance(r, dict))

        r = self.client.get_container_limit(self.client.container)
        cquota = r.values()[0]
        newquota = 2 * int(cquota)

        r = self.client.create_container(sizelimit=newquota)
        self.assertTrue(isinstance(r, dict))

        r = self.client.get_container_limit(self.client.container)
        xquota = int(r.values()[0])
        self.assertEqual(newquota, xquota)

        r = self.client.create_container(versioning='auto')
        self.assertTrue(isinstance(r, dict))

        r = self.client.get_container_versioning(self.client.container)
        nvers = r.values()[0]
        self.assertEqual('auto', nvers)

        r = self.client.container_put(versioning='none')
        self.assertEqual(r.status_code, 202)

        r = self.client.get_container_versioning(self.client.container)
        nvers = r.values()[0]
        self.assertEqual('none', nvers)

        r = self.client.create_container(metadata={'m1': 'v1', 'm2': 'v2'})
        self.assertTrue(isinstance(r, dict))

        r = self.client.get_container_meta(self.client.container)
        self.assertTrue('x-container-meta-m1' in r)
        self.assertEqual(r['x-container-meta-m1'], 'v1')
        self.assertTrue('x-container-meta-m2' in r)
        self.assertEqual(r['x-container-meta-m2'], 'v2')

        r = self.client.container_put(metadata={'m1': '', 'm2': 'v2a'})
        self.assertEqual(r.status_code, 202)

        r = self.client.get_container_meta(self.client.container)
        self.assertTrue('x-container-meta-m1' not in r)
        self.assertTrue('x-container-meta-m2' in r)
        self.assertEqual(r['x-container-meta-m2'], 'v2a')

        self.client.del_container_meta(self.client.container)

    def test_container_post(self):
        """Test container_POST"""
        self._test_0060_container_post()

    def _test_0060_container_post(self):
        self.client.container = self.c2

        """Simple post"""
        r = self.client.container_post()
        self.assertEqual(r.status_code, 202)

        """post meta"""
        self.client.set_container_meta({'m1': 'v1', 'm2': 'v2'})
        r = self.client.get_container_meta(self.client.container)
        self.assertTrue('x-container-meta-m1' in r)
        self.assertEqual(r['x-container-meta-m1'], 'v1')
        self.assertTrue('x-container-meta-m2' in r)
        self.assertEqual(r['x-container-meta-m2'], 'v2')

        """post/2del meta"""
        r = self.client.del_container_meta('m1')
        r = self.client.set_container_meta({'m2': 'v2a'})
        r = self.client.get_container_meta(self.client.container)
        self.assertTrue('x-container-meta-m1' not in r)
        self.assertTrue('x-container-meta-m2' in r)
        self.assertEqual(r['x-container-meta-m2'], 'v2a')

        """check quota"""
        r = self.client.get_container_limit(self.client.container)
        cquota = r.values()[0]
        newquota = 2 * int(cquota)
        r = self.client.set_container_limit(newquota)
        r = self.client.get_container_limit(self.client.container)
        xquota = int(r.values()[0])
        self.assertEqual(newquota, xquota)
        r = self.client.set_container_limit(cquota)
        r = self.client.get_container_limit(self.client.container)
        xquota = r.values()[0]
        self.assertEqual(cquota, xquota)

        """Check versioning"""
        self.client.set_container_versioning('auto')
        r = self.client.get_container_versioning(self.client.container)
        nvers = r.values()[0]
        self.assertEqual('auto', nvers)
        self.client.set_container_versioning('none')
        r = self.client.get_container_versioning(self.client.container)
        nvers = r.values()[0]
        self.assertEqual('none', nvers)

        """put_block uses content_type and content_length to
        post blocks of data 2 container. All that in upload_object"""
        """Change a file at fs"""
        f = self.create_large_file(1024 * 1024 * 100)
        """Upload it at a directory in container"""
        self.client.create_directory('dir')
        r = self.client.upload_object('/dir/sample.file', f)
        for term in ('content-length', 'content-type', 'x-object-version'):
            self.assertTrue(term in r)
        """Check if file has been uploaded"""
        r = self.client.get_object_info('/dir/sample.file')
        self.assertTrue(int(r['content-length']) > 100000000)

        """What is tranfer_encoding? What should I check about it? """
        #TODO

        """Check update=False"""
        r = self.client.object_post(
            'test',
            update=False,
            metadata={'newmeta': 'newval'})

        r = self.client.get_object_info('test')
        self.assertTrue('x-object-meta-newmeta' in r)
        self.assertFalse('x-object-meta-incontainer' in r)

        r = self.client.del_container_meta('m2')

    def test_container_delete(self):
        """Test container_DELETE"""
        self._test_0070_container_delete()

    def _test_0070_container_delete(self):

        """Fail to delete a non-empty container"""
        self.client.container = self.c2
        r = self.client.container_delete(success=409)
        self.assertEqual(r.status_code, 409)

        """Fail to delete c3 (empty) container"""
        self.client.container = self.c3
        r = self.client.container_delete(until='1000000000')
        self.assertEqual(r.status_code, 204)

        """Delete c3 (empty) container"""
        r = self.client.container_delete()
        self.assertEqual(r.status_code, 204)

        """Purge container(empty a container), check versionlist"""
        self.client.container = self.c1
        r = self.client.object_head('test', success=(200, 404))
        self.assertEqual(r.status_code, 200)
        self.client.del_container(delimiter='/')
        r = self.client.object_head('test', success=(200, 404))
        self.assertEqual(r.status_code, 404)
        r = self.client.get_object_versionlist('test')
        self.assertTrue(len(r) > 0)
        self.assertTrue(len(r[0]) > 1)
        self.client.purge_container()
        self.assertRaises(
            ClientError, self.client.get_object_versionlist, 'test')

    def _test_0080_recreate_deleted_data(self):
        self._init_data()

    def test_object_head(self):
        """Test object_HEAD"""
        self._test_0090_object_head()

    def _test_0090_object_head(self):
        self.client.container = self.c2
        obj = 'test'

        r = self.client.object_head(obj)
        self.assertEqual(r.status_code, 200)
        etag = r.headers['etag']
        real_version = r.headers['x-object-version']

        self.assertRaises(
            ClientError, self.client.object_head, obj, version=-10)
        r = self.client.object_head(obj, version=real_version)
        self.assertEqual(r.headers['x-object-version'], real_version)

        r = self.client.object_head(obj, if_etag_match=etag)
        self.assertEqual(r.status_code, 200)

        r = self.client.object_head(
            obj, if_etag_not_match=etag, success=(200, 412, 304))
        self.assertNotEqual(r.status_code, 200)

        r = self.client.object_head(
            obj, version=real_version, if_etag_match=etag, success=200)
        self.assertEqual(r.status_code, 200)

        """Check and if(un)modified_since"""
        for format in self.client.DATE_FORMATS:
            now_formated = self.now_unformated.strftime(format)
            r1 = self.client.object_head(
                obj, if_modified_since=now_formated, success=(200, 304, 412))
            sc1 = r1.status_code
            r2 = self.client.object_head(
                obj, if_unmodified_since=now_formated, success=(200, 304, 412))
            sc2 = r2.status_code
            self.assertNotEqual(sc1, sc2)

    def test_object_get(self):
        """Test object_GET"""
        self._test_0100_object_get()

    def _test_0100_object_get(self):
        self.client.container = self.c1
        obj = 'test'

        r = self.client.object_get(obj)
        self.assertEqual(r.status_code, 200)

        osize = int(r.headers['content-length'])
        etag = r.headers['etag']

        r = self.client.object_get(obj, hashmap=True)
        for term in ('hashes', 'block_hash', 'block_hash', 'bytes'):
            self.assertTrue(term in r.json)

        r = self.client.object_get(obj, format='xml', hashmap=True)
        self.assertEqual(len(r.text.split('hash>')), 3)

        rangestr = 'bytes=%s-%s' % (osize / 3, osize / 2)
        r = self.client.object_get(
            obj, data_range=rangestr, success=(200, 206))
        partsize = int(r.headers['content-length'])
        self.assertTrue(0 < partsize and partsize <= 1 + osize / 3)

        rangestr = 'bytes=%s-%s' % (osize / 3, osize / 2)
        r = self.client.object_get(
            obj, data_range=rangestr, if_range=True, success=(200, 206))
        partsize = int(r.headers['content-length'])
        self.assertTrue(0 < partsize and partsize <= 1 + osize / 3)

        r = self.client.object_get(obj, if_etag_match=etag)
        self.assertEqual(r.status_code, 200)

        r = self.client.object_get(obj, if_etag_not_match=etag + 'LALALA')
        self.assertEqual(r.status_code, 200)

        """Check and if(un)modified_since"""
        for format in self.client.DATE_FORMATS:
            now_formated = self.now_unformated.strftime(format)
            r1 = self.client.object_get(
                obj, if_modified_since=now_formated, success=(200, 304, 412))
            sc1 = r1.status_code
            r2 = self.client.object_get(
                obj, if_unmodified_since=now_formated, success=(200, 304, 412))
            sc2 = r2.status_code
            self.assertNotEqual(sc1, sc2)

        """Upload an object to download"""
        container_info_cache = dict()
        trg_fname = 'remotefile_%s' % self.now
        f_size = 59247824
        src_f = self.create_large_file(f_size)
        print('\tUploading...')
        r = self.client.upload_object(
            trg_fname, src_f, container_info_cache=container_info_cache)
        print('\tDownloading...')
        self.files.append(NamedTemporaryFile())
        dnl_f = self.files[-1]
        self.client.download_object(trg_fname, dnl_f)

        print('\tCheck if files match...')
        for pos in (0, f_size / 2, f_size - 128):
            src_f.seek(pos)
            dnl_f.seek(pos)
            self.assertEqual(src_f.read(64), dnl_f.read(64))

        print('\tDownload KiBs to string and check again...')
        for pos in (0, f_size / 2, f_size - 256):
            src_f.seek(pos)
            tmp_s = self.client.download_to_string(
                trg_fname, range_str='%s-%s' % (pos, (pos + 128)))
            self.assertEqual(tmp_s, src_f.read(len(tmp_s)))
        print('\tUploading KiBs as strings...')
        trg_fname = 'fromString_%s' % self.now
        src_size = 2 * 1024
        src_f.seek(0)
        src_str = src_f.read(src_size)
        self.client.upload_from_string(trg_fname, src_str)
        print('\tDownload as string and check...')
        tmp_s = self.client.download_to_string(trg_fname)
        self.assertEqual(tmp_s, src_str)

        """Upload a boring file"""
        trg_fname = 'boringfile_%s' % self.now
        src_f = self.create_boring_file(42)
        print('\tUploading boring file...')
        self.client.upload_object(
            trg_fname, src_f, container_info_cache=container_info_cache)
        print('\tDownloading boring file...')
        self.files.append(NamedTemporaryFile())
        dnl_f = self.files[-1]
        self.client.download_object(trg_fname, dnl_f)

        print('\tCheck if files match...')
        for i in range(42):
            self.assertEqual(sample_block(src_f, i), sample_block(dnl_f, i))

    def test_object_put(self):
        """Test object_PUT"""
        self._test_0150_object_put()

    def _test_0150_object_put(self):
        self.client.container = self.c2
        obj = 'another.test'

        self.client.create_object(obj + '.FAKE')
        r = self.client.get_object_info(obj + '.FAKE')
        self.assertEqual(r['content-type'], 'application/octet-stream')

        """create the object"""
        r = self.client.object_put(
            obj,
            data='a',
            content_type='application/octer-stream',
            permissions=dict(
                read=['accX:groupA', 'u1', 'u2'],
                write=['u2', 'u3']),
            metadata=dict(key1='val1', key2='val2'),
            content_encoding='UTF-8',
            content_disposition='attachment; filename="fname.ext"')
        self.assertEqual(r.status_code, 201)
        etag = r.headers['etag']

        """Check content-disposition"""
        r = self.client.get_object_info(obj)
        self.assertTrue('content-disposition' in r)

        """Check permissions"""
        r = self.client.get_object_sharing(obj)
        self.assertTrue('accx:groupa' in r['read'])
        self.assertTrue('u1' in r['read'])
        self.assertTrue('u2' in r['write'])
        self.assertTrue('u3' in r['write'])

        """Check metadata"""
        r = self.client.get_object_meta(obj)
        self.assertEqual(r['x-object-meta-key1'], 'val1')
        self.assertEqual(r['x-object-meta-key2'], 'val2')

        """Check public and if_etag_match"""
        r = self.client.object_put(
            obj,
            if_etag_match=etag,
            data='b',
            content_type='application/octet-stream',
            public=True)

        r = self.client.object_get(obj)
        self.assertTrue('x-object-public' in r.headers)
        vers2 = int(r.headers['x-object-version'])
        etag = r.headers['etag']
        self.assertEqual(r.text, 'b')

        """Check if_etag_not_match"""
        r = self.client.object_put(
            obj,
            if_etag_not_match=etag,
            data='c',
            content_type='application/octet-stream',
            success=(201, 412))
        self.assertEqual(r.status_code, 412)

        """Check content_type and content_length"""
        tmpdir = 'dir' + unicode(self.now)
        r = self.client.object_put(
            tmpdir, content_type='application/directory', content_length=0)

        r = self.client.get_object_info(tmpdir)
        self.assertEqual(r['content-type'], 'application/directory')

        """Check copy_from, content_encoding"""
        r = self.client.object_put(
            '%s/%s' % (tmpdir, obj),
            format=None,
            copy_from='/%s/%s' % (self.client.container, obj),
            content_encoding='application/octet-stream',
            source_account=self.client.account,
            content_length=0,
            success=201)
        self.assertEqual(r.status_code, 201)

        """Test copy_object for cross-conctainer copy"""
        self.client.copy_object(
            src_container=self.c2,
            src_object='%s/%s' % (tmpdir, obj),
            dst_container=self.c1,
            dst_object=obj)
        self.client.container = self.c1
        r1 = self.client.get_object_info(obj)
        self.client.container = self.c2
        r2 = self.client.get_object_info('%s/%s' % (tmpdir, obj))
        self.assertEqual(r1['x-object-hash'], r2['x-object-hash'])

        """Check cross-container copy_from, content_encoding"""
        self.client.container = self.c1
        fromstr = '/%s/%s/%s' % (self.c2, tmpdir, obj)
        r = self.client.object_put(
            obj,
            format=None,
            copy_from=fromstr,
            content_encoding='application/octet-stream',
            source_account=self.client.account,
            content_length=0,
            success=201)

        self.assertEqual(r.status_code, 201)
        r = self.client.get_object_info(obj)
        self.assertEqual(r['etag'], etag)

        """Check source_account"""
        self.client.container = self.c2
        fromstr = '/%s/%s' % (self.c1, obj)
        r = self.client.object_put(
            '%sv2' % obj,
            format=None,
            move_from=fromstr,
            content_encoding='application/octet-stream',
            source_account='nonExistendAddress@NeverLand.com',
            content_length=0,
            success=(201, 403))
        self.assertEqual(r.status_code, 403)

        """Check cross-container move_from"""
        self.client.container = self.c1
        r1 = self.client.get_object_info(obj)
        self.client.container = self.c2
        self.client.move_object(
            src_container=self.c1,
            src_object=obj,
            dst_container=self.c2,
            dst_object=obj + 'v0')
        r0 = self.client.get_object_info(obj + 'v0')
        self.assertEqual(r1['x-object-hash'], r0['x-object-hash'])

        """Check move_from"""
        r = self.client.object_put(
            '%sv1' % obj,
            format=None,
            move_from='/%s/%s' % (self.c2, obj),
            source_version=vers2,
            content_encoding='application/octet-stream',
            content_length=0, success=201)

        """Check manifest"""
        mobj = 'manifest.test'
        txt = ''
        for i in range(10):
            txt += '%s' % i
            r = self.client.object_put(
                '%s/%s' % (mobj, i),
                data='%s' % i,
                content_length=1,
                success=201,
                content_type='application/octet-stream',
                content_encoding='application/octet-stream')

        r = self.client.object_put(
            mobj,
            content_length=0,
            content_type='application/octet-stream',
            manifest='%s/%s' % (self.client.container, mobj))

        r = self.client.object_get(mobj)
        self.assertEqual(r.text, txt)

        """Upload a local file with one request"""
        newf = self.create_large_file(1024 * 10)
        self.client.upload_object('sample.file', newf)
        """Check if file has been uploaded"""
        r = self.client.get_object_info('sample.file')
        self.assertEqual(int(r['content-length']), 10240)

        """Some problems with transfer-encoding?"""

    def test_object_copy(self):
        """Test object_COPY"""
        self._test_0110_object_copy()

    def _test_0110_object_copy(self):
        #  TODO: check with source_account option
        self.client.container = self.c2
        obj = 'test2'

        data = '{"key1":"val1", "key2":"val2"}'
        r = self.client.object_put(
            '%sorig' % obj,
            content_type='application/octet-stream',
            data=data,
            metadata=dict(mkey1='mval1', mkey2='mval2'),
            permissions=dict(
                read=['accX:groupA', 'u1', 'u2'],
                write=['u2', 'u3']),
            content_disposition='attachment; filename="fname.ext"')

        r = self.client.object_copy(
            '%sorig' % obj,
            destination='/%s/%s' % (self.client.container, obj),
            ignore_content_type=False, content_type='application/json',
            metadata={'mkey2': 'mval2a', 'mkey3': 'mval3'},
            permissions={'write': ['u5', 'accX:groupB']})
        self.assertEqual(r.status_code, 201)

        """Check content-disposition"""
        r = self.client.get_object_info(obj)
        self.assertTrue('content-disposition' in r)

        """Check Metadata"""
        r = self.client.get_object_meta(obj)
        self.assertEqual(r['x-object-meta-mkey1'], 'mval1')
        self.assertEqual(r['x-object-meta-mkey2'], 'mval2a')
        self.assertEqual(r['x-object-meta-mkey3'], 'mval3')

        """Check permissions"""
        r = self.client.get_object_sharing(obj)
        self.assertFalse('read' in r or 'u2' in r['write'])
        self.assertTrue('accx:groupb' in r['write'])

        """Check destination account"""
        r = self.client.object_copy(
            obj,
            destination='/%s/%s' % (self.c1, obj),
            content_encoding='utf8',
            content_type='application/json',
            destination_account='nonExistendAddress@NeverLand.com',
            success=(201, 404))
        self.assertEqual(r.status_code, 404)

        """Check destination being another container
        and also content_type and content encoding"""
        r = self.client.object_copy(
            obj,
            destination='/%s/%s' % (self.c1, obj),
            content_encoding='utf8',
            content_type='application/json')
        self.assertEqual(r.status_code, 201)
        self.assertEqual(
            r.headers['content-type'],
            'application/json; charset=UTF-8')

        """Check ignore_content_type and content_type"""
        r = self.client.object_get(obj)
        etag = r.headers['etag']
        ctype = r.headers['content-type']
        self.assertEqual(ctype, 'application/json')

        r = self.client.object_copy(
            '%sorig' % obj,
            destination='/%s/%s0' % (self.client.container, obj),
            ignore_content_type=True,
            content_type='application/json')
        self.assertEqual(r.status_code, 201)
        self.assertNotEqual(r.headers['content-type'], 'application/json')

        """Check if_etag_(not_)match"""
        r = self.client.object_copy(
            obj,
            destination='/%s/%s1' % (self.client.container, obj),
            if_etag_match=etag)
        self.assertEqual(r.status_code, 201)

        r = self.client.object_copy(
            obj,
            destination='/%s/%s2' % (self.client.container, obj),
            if_etag_not_match='lalala')
        self.assertEqual(r.status_code, 201)
        vers2 = r.headers['x-object-version']

        """Check source_version, public and format """
        r = self.client.object_copy(
            '%s2' % obj,
            destination='/%s/%s3' % (self.client.container, obj),
            source_version=vers2,
            format='xml',
            public=True)
        self.assertEqual(r.status_code, 201)
        self.assertTrue(r.headers['content-type'].index('xml') > 0)

        r = self.client.get_object_info(obj + '3')
        self.assertTrue('x-object-public' in r)

    def test_object_move(self):
        """Test object_MOVE"""
        self._test_0120_object_move()

    def _test_0120_object_move(self):
        self.client.container = self.c2
        obj = 'test2'

        data = '{"key1": "val1", "key2": "val2"}'
        r = self.client.object_put(
            '%sorig' % obj,
            content_type='application/octet-stream',
            data=data,
            metadata=dict(mkey1='mval1', mkey2='mval2'),
            permissions=dict(
                read=['accX:groupA', 'u1', 'u2'],
                write=['u2', 'u3']))

        r = self.client.object_move(
            '%sorig' % obj,
            destination='/%s/%s' % (self.client.container, obj),
            ignore_content_type=False,
            content_type='application/json',
            metadata=dict(mkey2='mval2a', mkey3='mval3'),
            permissions=dict(write=['u5', 'accX:groupB']))
        self.assertEqual(r.status_code, 201)

        """Check Metadata"""
        r = self.client.get_object_meta(obj)
        self.assertEqual(r['x-object-meta-mkey1'], 'mval1')
        self.assertEqual(r['x-object-meta-mkey2'], 'mval2a')
        self.assertEqual(r['x-object-meta-mkey3'], 'mval3')

        """Check permissions"""
        r = self.client.get_object_sharing(obj)
        self.assertFalse('read' in r)
        self.assertTrue('u5' in r['write'])
        self.assertTrue('accx:groupb' in r['write'])

        """Check destination account"""
        r = self.client.object_move(
            obj,
            destination='/%s/%s' % (self.c1, obj),
            content_encoding='utf8',
            content_type='application/json',
            destination_account='nonExistendAddress@NeverLand.com',
            success=(201, 404))
        self.assertEqual(r.status_code, 404)

        """Check destination being another container and also
        content_type, content_disposition and content encoding"""
        r = self.client.object_move(
            obj,
            destination='/%s/%s' % (self.c1, obj),
            content_encoding='utf8',
            content_type='application/json',
            content_disposition='attachment; filename="fname.ext"')
        self.assertEqual(r.status_code, 201)
        self.assertEqual(
            r.headers['content-type'],
            'application/json; charset=UTF-8')
        self.client.container = self.c1
        r = self.client.get_object_info(obj)
        self.assertTrue('content-disposition' in r)
        self.assertTrue('fname.ext' in r['content-disposition'])
        etag = r['etag']
        ctype = r['content-type']
        self.assertEqual(ctype, 'application/json')

        """Check ignore_content_type and content_type"""
        r = self.client.object_move(
            obj,
            destination='/%s/%s' % (self.c2, obj),
            ignore_content_type=True,
            content_type='application/json')
        self.assertEqual(r.status_code, 201)
        self.assertNotEqual(r.headers['content-type'], 'application/json')

        """Check if_etag_(not_)match"""
        self.client.container = self.c2
        r = self.client.object_move(
            obj,
            destination='/%s/%s0' % (self.client.container, obj),
            if_etag_match=etag)
        self.assertEqual(r.status_code, 201)

        r = self.client.object_move(
            '%s0' % obj,
            destination='/%s/%s1' % (self.client.container, obj),
            if_etag_not_match='lalala')
        self.assertEqual(r.status_code, 201)

        """Check public and format """
        r = self.client.object_move(
            '%s1' % obj,
            destination='/%s/%s2' % (self.client.container, obj),
            format='xml',
            public=True)
        self.assertEqual(r.status_code, 201)
        self.assertTrue(r.headers['content-type'].index('xml') > 0)

        r = self.client.get_object_info(obj + '2')
        self.assertTrue('x-object-public' in r)

    def test_object_post(self):
        """Test object_POST"""
        self._test_0130_object_post()

    def _test_0130_object_post(self):
        self.client.container = self.c2
        obj = 'test2'

        """create a filesystem file"""
        self.files.append(NamedTemporaryFile())
        newf = self.files[-1]
        newf.writelines([
            'ello!\n',
            'This is a test line\n',
            'inside a test file\n'])

        """create a file on container"""
        r = self.client.object_put(
            obj,
            content_type='application/octet-stream',
            data='H',
            metadata=dict(mkey1='mval1', mkey2='mval2'),
            permissions=dict(
                read=['accX:groupA', 'u1', 'u2'],
                write=['u2', 'u3']))

        """Append livetest update, content_[range|type|length]"""
        newf.seek(0)
        self.client.append_object(obj, newf)
        r = self.client.object_get(obj)
        self.assertTrue(r.text.startswith('Hello!'))

        """Overwrite livetest update,
            content_type, content_length, content_range
        """
        newf.seek(0)
        r = self.client.overwrite_object(obj, 0, 10, newf)
        r = self.client.object_get(obj)
        self.assertTrue(r.text.startswith('ello!'))

        """Truncate livetest update,
            content_range, content_type, object_bytes and source_object"""
        r = self.client.truncate_object(obj, 5)
        r = self.client.object_get(obj)
        self.assertEqual(r.text, 'ello!')

        """Check metadata"""
        self.client.set_object_meta(obj, {'mkey2': 'mval2a', 'mkey3': 'mval3'})
        r = self.client.get_object_meta(obj)
        self.assertEqual(r['x-object-meta-mkey1'], 'mval1')
        self.assertEqual(r['x-object-meta-mkey2'], 'mval2a')
        self.assertEqual(r['x-object-meta-mkey3'], 'mval3')
        self.client.del_object_meta(obj, 'mkey1')
        r = self.client.get_object_meta(obj)
        self.assertFalse('x-object-meta-mkey1' in r)

        """Check permissions"""
        self.client.set_object_sharing(
            obj, read_permission=['u4', 'u5'], write_permission=['u4'])
        r = self.client.get_object_sharing(obj)
        self.assertTrue('read' in r)
        self.assertTrue('u5' in r['read'])
        self.assertTrue('write' in r)
        self.assertTrue('u4' in r['write'])
        self.client.del_object_sharing(obj)
        r = self.client.get_object_sharing(obj)
        self.assertTrue(len(r) == 0)

        """Check publish"""
        self.client.publish_object(obj)
        r = self.client.get_object_info(obj)
        self.assertTrue('x-object-public' in r)
        self.client.unpublish_object(obj)
        r = self.client.get_object_info(obj)
        self.assertFalse('x-object-public' in r)

        """Check if_etag_(not)match"""
        etag = r['etag']
        r = self.client.object_post(
            obj,
            update=True,
            public=True,
            if_etag_not_match=etag,
            success=(412, 202, 204))
        #self.assertEqual(r.status_code, 412)

        r = self.client.object_post(
            obj,
            update=True,
            public=True,
            if_etag_match=etag,
            content_encoding='application/json')

        r = self.client.get_object_info(obj)
        helloVersion = r['x-object-version']
        self.assertTrue('x-object-public' in r)
        self.assertEqual(r['content-encoding'], 'application/json')

        """Check source_version and source_account and content_disposition"""
        r = self.client.object_post(
            obj,
            update=True,
            content_type='application/octet-srteam',
            content_length=5,
            content_range='bytes 1-5/*',
            source_object='/%s/%s' % (self.c2, obj),
            source_account='thisAccountWillNeverExist@adminland.com',
            source_version=helloVersion,
            data='12345',
            success=(403, 202, 204))
        self.assertEqual(r.status_code, 403)

        r = self.client.object_post(
            obj,
            update=True,
            content_type='application/octet-srteam',
            content_length=5,
            content_range='bytes 1-5/*',
            source_object='/%s/%s' % (self.c2, obj),
            source_account=self.client.account,
            source_version=helloVersion,
            data='12345',
            content_disposition='attachment; filename="fname.ext"')

        r = self.client.object_get(obj)
        self.assertEqual(r.text, 'eello!')
        self.assertTrue('content-disposition' in r.headers)
        self.assertTrue('fname.ext' in r.headers['content-disposition'])

        """Check manifest"""
        mobj = 'manifest.test'
        txt = ''
        for i in range(10):
            txt += '%s' % i
            r = self.client.object_put(
                '%s/%s' % (mobj, i),
                data='%s' % i,
                content_length=1,
                success=201,
                content_encoding='application/octet-stream',
                content_type='application/octet-stream')

        self.client.create_object_by_manifestation(
            mobj, content_type='application/octet-stream')

        r = self.client.object_post(
            mobj, manifest='%s/%s' % (self.client.container, mobj))

        r = self.client.object_get(mobj)
        self.assertEqual(r.text, txt)

        """We need to check transfer_encoding """

    def test_object_delete(self):
        """Test object_DELETE"""
        self._test_0140_object_delete()

    def _test_0140_object_delete(self):
        self.client.container = self.c2
        obj = 'test2'
        """create a file on container"""
        r = self.client.object_put(
            obj,
            content_type='application/octet-stream',
            data='H',
            metadata=dict(mkey1='mval1', mkey2='mval2'),
            permissions=dict(
                read=['accX:groupA', 'u1', 'u2'],
                write=['u2', 'u3']))

        """Check with false until"""
        r = self.client.object_delete(obj, until=1000000)

        r = self.client.object_get(obj, success=(200, 404))
        self.assertEqual(r.status_code, 200)

        """Check normal case"""
        r = self.client.object_delete(obj)
        self.assertEqual(r.status_code, 204)

        r = self.client.object_get(obj, success=(200, 404))
        self.assertEqual(r.status_code, 404)

    def create_large_file(self, size):
        """Create a large file at fs"""
        print
        self.files.append(NamedTemporaryFile())
        f = self.files[-1]
        Ki = size / 8
        bytelist = [b * Ki for b in range(size / Ki)]

        def append2file(step):
            f.seek(step)
            f.write(urandom(Ki))
            f.flush()
        self.do_with_progress_bar(
            append2file,
            ' create rand file %s (%sB): ' % (f.name, size),
            bytelist)
        f.seek(0)
        return f

    def create_boring_file(self, num_of_blocks):
        """Create a file with some blocks being the same"""
        self.files.append(NamedTemporaryFile())
        tmpFile = self.files[-1]
        block_size = 4 * 1024 * 1024
        print('\n\tCreate boring file of %s blocks' % num_of_blocks)
        chars = chargen()
        while num_of_blocks:
            fslice = 3 if num_of_blocks > 3 else num_of_blocks
            tmpFile.write(fslice * block_size * chars.next())
            num_of_blocks -= fslice
        print('\t\tDone')
        tmpFile.seek(0)
        return tmpFile



# Copyright 2011-2013 GRNET S.A. All rights reserved.
#
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
#
#   1. Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and
# documentation are those of the authors and should not be
# interpreted as representing official policies, either expressed
# or implied, of GRNET S.A.command

from traceback import print_stack, print_exc

from kamaki.clients import ClientError
from kamaki.cli.errors import CLIError, raiseCLIError, CLISyntaxError
from kamaki.cli import _debug, kloger
from kamaki.cli.utils import format_size

CLOUDNAME = [
    'Note: If you use a named cloud, use its name instead of "default"']


class generic(object):

    @classmethod
    def all(this, foo):
        def _raise(self, *args, **kwargs):
            try:
                return foo(self, *args, **kwargs)
            except Exception as e:
                if _debug:
                    print_stack()
                    print_exc(e)
                if isinstance(e, CLIError) or isinstance(e, ClientError):
                    raiseCLIError(e)
                raiseCLIError(e, details=['%s, -d for debug info' % type(e)])
        return _raise

    @classmethod
    def _connection(this, foo):
        def _raise(self, *args, **kwargs):
            try:
                foo(self, *args, **kwargs)
            except ClientError as ce:
                ce_msg = ('%s' % ce).lower()
                if ce.status == 401:
                    raiseCLIError(ce, 'Authorization failed', details=[
                        'Make sure a valid token is provided:',
                        '  to check if token is valid: /user authenticate',
                        '  to set token:',
                        '    /config set cloud.default.token <token>',
                        '  to get current token:',
                        '    /config get cloud.default.token'] + CLOUDNAME)
                elif ce.status in range(-12, 200) + [302, 401, 403, 500]:
                    raiseCLIError(ce, importance=3, details=[
                        'Check if service is up'])
                elif ce.status == 404 and 'kamakihttpresponse' in ce_msg:
                    client = getattr(self, 'client', None)
                    if not client:
                        raise
                    url = getattr(client, 'base_url', '<empty>')
                    msg = 'Invalid service URL %s' % url
                    raiseCLIError(ce, msg, details=[
                        'Check if authentication URL is correct',
                        '  check current URL:',
                        '    /config get cloud.default.url',
                        '  set new authentication URL:',
                        '    /config set cloud.default.url'] + CLOUDNAME)
                raise
        return _raise


class user(object):

    _token_details = [
        'To check default token: /config get cloud.default.token',
        'If set/update a token:',
        '*  (permanent):  /config set cloud.default.token <token>',
        '*  (temporary):  re-run with <token> parameter'] + CLOUDNAME

    @classmethod
    def load(this, foo):
        def _raise(self, *args, **kwargs):
            r = foo(self, *args, **kwargs)
            try:
                client = getattr(self, 'client')
            except AttributeError as ae:
                raiseCLIError(ae, 'Client setup failure', importance=3)
            if not getattr(client, 'token', False):
                kloger.warning(
                    'No permanent token (try:'
                    ' kamaki config set cloud.default.token <tkn>)')
            if not getattr(client, 'base_url', False):
                msg = 'Missing synnefo authentication URL'
                raise CLIError(msg, importance=3, details=[
                    'Check if authentication URL is correct',
                        '  check current URL:',
                        '    /config get cloud.default.url',
                        '  set new auth. URL:',
                        '    /config set cloud.default.url'] + CLOUDNAME)
            return r
        return _raise

    @classmethod
    def authenticate(this, foo):
        def _raise(self, *args, **kwargs):
            try:
                return foo(self, *args, **kwargs)
            except ClientError as ce:
                if ce.status == 401:
                    token = kwargs.get('custom_token', 0) or self.client.token
                    msg = (
                        'Authorization failed for token %s' % token
                    ) if token else 'No token provided',
                    details = [] if token else this._token_details
                    raiseCLIError(ce, msg, details=details)
                raise ce
            self._raise = foo
        return _raise


class history(object):
    @classmethod
    def init(this, foo):
        def _raise(self, *args, **kwargs):
            r = foo(self, *args, **kwargs)
            if not hasattr(self, 'history'):
                raise CLIError('Failed to load history', importance=2)
            return r
        return _raise

    @classmethod
    def _get_cmd_ids(this, foo):
        def _raise(self, cmd_ids, *args, **kwargs):
            if not cmd_ids:
                raise CLISyntaxError(
                    'Usage: <id1|id1-id2> [id3|id3-id4] ...',
                    details=self.__doc__.split('\n'))
            return foo(self, cmd_ids, *args, **kwargs)
        return _raise


class cyclades(object):
    about_flavor_id = [
        'How to pick a valid flavor id:',
        '* get a list of flavor ids: /flavor list',
        '* details of flavor: /flavor info <flavor id>']

    about_network_id = [
        'How to pick a valid network id:',
        '* get a list of network ids: /network list',
        '* details of network: /network info <network id>']

    @classmethod
    def connection(this, foo):
        return generic._connection(foo)

    @classmethod
    def date(this, foo):
        def _raise(self, *args, **kwargs):
            try:
                return foo(self, *args, **kwargs)
            except ClientError as ce:
                if ce.status == 400 and 'changes-since' in ('%s' % ce):
                    raise CLIError(
                        'Incorrect date format for --since',
                        details=['Accepted date format: d/m/y'])
                raise
        return _raise

    @classmethod
    def cluster_size(this, foo):
        def _raise(self, *args, **kwargs):
            size = kwargs.get('size', None)
            try:
                size = int(size)
                assert size > 0, 'Cluster size must be a positive integer'
                return foo(self, *args, **kwargs)
            except ValueError as ve:
                msg = 'Invalid cluster size value %s' % size
                raiseCLIError(ve, msg, importance=1, details=[
                    'Cluster size must be a positive integer'])
            except AssertionError as ae:
                raiseCLIError(
                    ae, 'Invalid cluster size %s' % size, importance=1)
            except ClientError:
                raise
        return _raise

    @classmethod
    def network_id(this, foo):
        def _raise(self, *args, **kwargs):
            network_id = kwargs.get('network_id', None)
            try:
                network_id = int(network_id)
                return foo(self, *args, **kwargs)
            except ValueError as ve:
                msg = 'Invalid network id %s ' % network_id
                details = 'network id must be a positive integer'
                raiseCLIError(ve, msg, details=details, importance=1)
            except ClientError as ce:
                if network_id and ce.status == 404 and (
                    'network' in ('%s' % ce).lower()
                ):
                    msg = 'No network with id %s found' % network_id,
                    raiseCLIError(ce, msg, details=this.about_network_id)
                raise
        return _raise

    @classmethod
    def network_max(this, foo):
        def _raise(self, *args, **kwargs):
            try:
                return foo(self, *args, **kwargs)
            except ClientError as ce:
                if ce.status == 413:
                    msg = 'Cannot create another network',
                    details = [
                        'Maximum number of networks reached',
                        '* to get a list of networks: /network list',
                        '* to delete a network: /network delete <net id>']
                    raiseCLIError(ce, msg, details=details)
                raise
        return _raise

    @classmethod
    def network_in_use(this, foo):
        def _raise(self, *args, **kwargs):
            network_id = kwargs.get('network_id', None)
            try:
                return foo(self, *args, **kwargs)
            except ClientError as ce:
                if network_id and ce.status in (400, ):
                    msg = 'Network with id %s does not exist' % network_id,
                    raiseCLIError(ce, msg, details=this.about_network_id)
                elif network_id or ce.status in (421, ):
                    msg = 'Network with id %s is in use' % network_id,
                    raiseCLIError(ce, msg, details=[
                        'Disconnect all nics/servers of this network first',
                        '* to get nics: /network info %s' % network_id,
                        '.  (under "attachments" section)',
                        '* to disconnect: /network disconnect <nic id>'])
                raise
        return _raise

    @classmethod
    def flavor_id(this, foo):
        def _raise(self, *args, **kwargs):
            flavor_id = kwargs.get('flavor_id', None)
            try:
                flavor_id = int(flavor_id)
                return foo(self, *args, **kwargs)
            except ValueError as ve:
                msg = 'Invalid flavor id %s ' % flavor_id,
                details = 'Flavor id must be a positive integer'
                raiseCLIError(ve, msg, details=details, importance=1)
            except ClientError as ce:
                if flavor_id and ce.status == 404 and (
                    'flavor' in ('%s' % ce).lower()
                ):
                        msg = 'No flavor with id %s found' % flavor_id,
                        raiseCLIError(ce, msg, details=this.about_flavor_id)
                raise
        return _raise

    @classmethod
    def server_id(this, foo):
        def _raise(self, *args, **kwargs):
            server_id = kwargs.get('server_id', None)
            try:
                server_id = int(server_id)
                return foo(self, *args, **kwargs)
            except ValueError as ve:
                msg = 'Invalid virtual server id %s' % server_id,
                details = 'Server id must be a positive integer'
                raiseCLIError(ve, msg, details=details, importance=1)
            except ClientError as ce:
                err_msg = ('%s' % ce).lower()
                if (
                    ce.status == 404 and 'server' in err_msg
                ) or (
                    ce.status == 400 and 'not found' in err_msg
                ):
                    msg = 'virtual server with id %s not found' % server_id,
                    raiseCLIError(ce, msg, details=[
                        '* to get ids of all servers: /server list',
                        '* to get server details: /server info <server id>'])
                raise
        return _raise

    @classmethod
    def firewall(this, foo):
        def _raise(self, *args, **kwargs):
            profile = kwargs.get('profile', None)
            try:
                return foo(self, *args, **kwargs)
            except ClientError as ce:
                if ce.status == 400 and profile and (
                    'firewall' in ('%s' % ce).lower()
                ):
                    msg = '%s is an invalid firewall profile term' % profile
                    raiseCLIError(ce, msg, details=[
                        'Try one of the following:',
                        '* DISABLED: Shutdown firewall',
                        '* ENABLED: Firewall in normal mode',
                        '* PROTECTED: Firewall in secure mode'])
                raise
        return _raise

    @classmethod
    def nic_id(this, foo):
        def _raise(self, *args, **kwargs):
            try:
                return foo(self, *args, **kwargs)
            except ClientError as ce:
                nic_id = kwargs.get('nic_id', None)
                if nic_id and ce.status == 404 and (
                    'network interface' in ('%s' % ce).lower()
                ):
                    server_id = kwargs.get('server_id', '<no server>')
                    err_msg = 'No nic %s on virtual server with id %s' % (
                        nic_id,
                        server_id)
                    raiseCLIError(ce, err_msg, details=[
                        '* check v. server with id %s: /server info %s' % (
                            server_id,
                            server_id),
                        '* list nics for v. server with id %s:' % server_id,
                        '      /server addr %s' % server_id])
                raise
        return _raise

    @classmethod
    def nic_format(this, foo):
        def _raise(self, *args, **kwargs):
            try:
                return foo(self, *args, **kwargs)
            except IndexError as ie:
                nic_id = kwargs.get('nic_id', None)
                msg = 'Invalid format for network interface (nic) %s' % nic_id
                raiseCLIError(ie, msg, importance=1, details=[
                    'nid_id format: nic-<server id>-<nic id>',
                    '* get nics of a network: /network info <net id>',
                    '    (listed the "attachments" section)'])
        return _raise

    @classmethod
    def metadata(this, foo):
        def _raise(self, *args, **kwargs):
            key = kwargs.get('key', None)
            try:
                foo(self, *args, **kwargs)
            except ClientError as ce:
                if key and ce.status == 404 and (
                    'metadata' in ('%s' % ce).lower()
                ):
                        raiseCLIError(
                            ce, 'No v. server metadata with key %s' % key)
                raise
        return _raise


class plankton(object):

    about_image_id = [
        'How to pick a suitable image:',
        '* get a list of image ids: /image list',
        '* details of image: /image meta <image id>']

    @classmethod
    def connection(this, foo):
        return generic._connection(foo)

    @classmethod
    def id(this, foo):
        def _raise(self, *args, **kwargs):
            image_id = kwargs.get('image_id', None)
            try:
                foo(self, *args, **kwargs)
            except ClientError as ce:
                if image_id and (
                    ce.status == 404
                    or (
                        ce.status == 400
                        and 'image not found' in ('%s' % ce).lower())
                    or ce.status == 411
                ):
                        msg = 'No image with id %s found' % image_id
                        raiseCLIError(ce, msg, details=this.about_image_id)
                raise
        return _raise

    @classmethod
    def metadata(this, foo):
        def _raise(self, *args, **kwargs):
            key = kwargs.get('key', None)
            try:
                return foo(self, *args, **kwargs)
            except ClientError as ce:
                ce_msg = ('%s' % ce).lower()
                if ce.status == 404 or (
                        ce.status == 400 and 'metadata' in ce_msg):
                    msg = 'No properties with key %s in this image' % key
                    raiseCLIError(ce, msg)
                raise
        return _raise


class pithos(object):
    container_howto = [
        'To specify a container:',
        '  1. --container=<container> (temporary, overrides all)',
        '  2. Use the container:path format (temporary, overrides 3)',
        '  3. Set pithos_container variable (permanent)',
        '     /config set pithos_container <container>',
        'For a list of containers: /file list']

    @classmethod
    def connection(this, foo):
        return generic._connection(foo)

    @classmethod
    def account(this, foo):
        def _raise(self, *args, **kwargs):
            try:
                return foo(self, *args, **kwargs)
            except ClientError as ce:
                if ce.status == 403:
                    raiseCLIError(
                        ce,
                        'Invalid account credentials for this operation',
                        details=['Check user account settings'])
                raise
        return _raise

    @classmethod
    def quota(this, foo):
        def _raise(self, *args, **kwargs):
            try:
                return foo(self, *args, **kwargs)
            except ClientError as ce:
                if ce.status == 413:
                    raiseCLIError(ce, 'User quota exceeded', details=[
                        '* get quotas:',
                        '  * upper total limit:      /file quota',
                        '  * container limit:',
                        '    /file containerlimit get <container>',
                        '* set a higher container limit:',
                        '    /file containerlimit set <limit> <container>'])
                raise
        return _raise

    @classmethod
    def container(this, foo):
        def _raise(self, *args, **kwargs):
            dst_cont = kwargs.get('dst_cont', None)
            try:
                return foo(self, *args, **kwargs)
            except ClientError as ce:
                if ce.status == 404 and 'container' in ('%s' % ce).lower():
                        cont = ('%s or %s' % (
                            self.container,
                            dst_cont)) if dst_cont else self.container
                        msg = 'Is container %s in current account?' % (cont),
                        raiseCLIError(ce, msg, details=this.container_howto)
                raise
        return _raise

    @classmethod
    def local_path(this, foo):
        def _raise(self, *args, **kwargs):
            local_path = kwargs.get('local_path', '<None>')
            try:
                return foo(self, *args, **kwargs)
            except IOError as ioe:
                msg = 'Failed to access file %s' % local_path,
                raiseCLIError(ioe, msg, importance=2)
        return _raise

    @classmethod
    def object_path(this, foo):
        def _raise(self, *args, **kwargs):
            try:
                return foo(self, *args, **kwargs)
            except ClientError as ce:
                err_msg = ('%s' % ce).lower()
                if (
                    ce.status == 404 or ce.status == 500
                ) and 'object' in err_msg and 'not' in err_msg:
                    msg = 'No object %s in container %s' % (
                        self.path,
                        self.container)
                    raiseCLIError(ce, msg, details=this.container_howto)
                raise
        return _raise

    @classmethod
    def object_size(this, foo):
        def _raise(self, *args, **kwargs):
            size = kwargs.get('size', None)
            start = kwargs.get('start', 0)
            end = kwargs.get('end', 0)
            if size:
                try:
                    size = int(size)
                except ValueError as ve:
                    msg = 'Invalid file size %s ' % size
                    details = ['size must be a positive integer']
                    raiseCLIError(ve, msg, details=details, importance=1)
            else:
                try:
                    start = int(start)
                except ValueError as e:
                    msg = 'Invalid start value %s in range' % start,
                    details = ['size must be a positive integer'],
                    raiseCLIError(e, msg, details=details, importance=1)
                try:
                    end = int(end)
                except ValueError as e:
                    msg = 'Invalid end value %s in range' % end
                    details = ['size must be a positive integer']
                    raiseCLIError(e, msg, details=details, importance=1)
                if start > end:
                    raiseCLIError(
                        'Invalid range %s-%s' % (start, end),
                        details=['size must be a positive integer'],
                        importance=1)
                size = end - start
            try:
                return foo(self, *args, **kwargs)
            except ClientError as ce:
                err_msg = ('%s' % ce).lower()
                expected = 'object length is smaller than range length'
                if size and (
                    ce.status == 416 or (
                        ce.status == 400 and expected in err_msg)):
                    raiseCLIError(ce, 'Remote object %s:%s <= %s %s' % (
                        self.container, self.path, format_size(size),
                        ('(%sB)' % size) if size >= 1024 else ''))
                raise
        return _raise



# Copyright 2011-2013 GRNET S.A. All rights reserved.
#
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
#
#   1. Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and
# documentation are those of the authors and should not be
# interpreted as representing official policies, either expressed
# or implied, of GRNET S.A.command

from kamaki.cli import command
from kamaki.clients.astakos import AstakosClient
from kamaki.cli.commands import (
    _command_init, errors, _optional_json, addLogSettings)
from kamaki.cli.command_tree import CommandTree
from kamaki.cli.errors import CLIBaseUrlError, CLIError

user_cmds = CommandTree('user', 'Astakos API commands')
_commands = [user_cmds]


class _user_init(_command_init):

    def _write_main_token(self, token):
        tokens = self.config.get_cloud(self.cloud, 'token').split()
        if token in tokens:
            tokens.remove(token)
        tokens.insert(0, token)
        self.config.set_cloud(self.cloud, 'token', ' '.join(tokens))
        self.config.write()

    @errors.generic.all
    @errors.user.load
    @addLogSettings
    def _run(self):
        if getattr(self, 'cloud', False):
            base_url = self._custom_url('astakos')
            if base_url:
                token = self._custom_token(
                    'astakos') or self.config.get_cloud(self.cloud, 'token')
                token = token.split()[0] if ' ' in token else token
                self.client = AstakosClient(base_url=base_url, token=token)
                return
        else:
            self.cloud = 'default'
        if getattr(self, 'auth_base', False):
            self.client = self.auth_base
            return
        raise CLIBaseUrlError(service='astakos')

    def main(self):
        self._run()


@command(user_cmds)
class user_authenticate(_user_init, _optional_json):
    """Authenticate a user
    Get user information (e.g., unique account name) from token
    Token should be set in settings:
    *  check if a token is set    /config whoami cloud.default.token
    *  permanently set a token    /config set cloud.default.token <token>
    Token can also be provided as a parameter
    (In case of another named cloud, use its name instead of default)
    """

    @errors.generic.all
    @errors.user.authenticate
    def _run(self, custom_token=None):
        token_bu = self.client.token
        try:
            r = self.client.authenticate(custom_token)
            if (token_bu != self.client.token and self.ask_user(
                    'Permanently save token as cloud.%s.token ?' % (
                        self.cloud))):
                self._write_main_token(self.client.token)
        except Exception:
            #recover old token
            self.client.token = token_bu
            raise

        def _print_access(r, out):
            self.print_dict(r['access'], out=out)

        self._print(r, _print_access)

    def main(self, custom_token=None):
        super(self.__class__, self)._run()
        self._run(custom_token)


@command(user_cmds)
class user_list(_user_init, _optional_json):
    """List all authenticated users"""

    @errors.generic.all
    def _run(self, custom_token=None):
        self._print(self.client.list_users())

    def main(self):
        super(self.__class__, self)._run()
        self._run()


@command(user_cmds)
class user_whoami(_user_init, _optional_json):
    """Get current session user information"""

    @errors.generic.all
    def _run(self):
        self._print(self.client.user_info(), self.print_dict)

    def main(self):
        super(self.__class__, self)._run()
        self._run()


@command(user_cmds)
class user_set(_user_init, _optional_json):
    """Set session user by id
    To enrich your options, authenticate more users:
    /user authenticate <other user token>
    To list authenticated users
    /user list
    To get the current session user
    /user whoami
    """

    @errors.generic.all
    def _run(self, uuid):
        for user in self.client.list_users():
            if user.get('id', None) in (uuid,):
                ntoken = user['auth_token']
                if ntoken == self.client.token:
                    self.error('%s (%s) is already the session user' % (
                        self.client.user_term('name'),
                        self.client.user_term('id')))
                    return
                self.client.token = user['auth_token']
                self.error('Session user set to %s (%s)' % (
                        self.client.user_term('name'),
                        self.client.user_term('id')))
                if self.ask_user(
                        'Permanently make %s the main user?' % (
                            self.client.user_term('name'))):
                    self._write_main_token(self.client.token)
                return
        raise CLIError(
            'User with UUID %s not authenticated in current session' % uuid,
            details=[
                'To authenticate a user', '  /user authenticate <t0k3n>'])

    def main(self, uuid):
        super(self.__class__, self)._run()
        self._run(uuid)
